from __future__ import annotations

import copy
import warnings
from typing import Optional, Self

import numpy as np
from numpy.typing import NDArray
from scipy.optimize import newton
from typing_extensions import override

from relife.economic import AgeReplacementReward, ExponentialDiscounting, cost
from relife.lifetime_model import (
    AgeReplacementModel,
    FrozenAgeReplacementModel,
    FrozenLeftTruncatedModel,
    LeftTruncatedModel,
)
from relife.quadrature import legendre_quadrature

from ..lifetime_model.distribution import LifetimeDistribution
from ..lifetime_model.regression import FrozenLifetimeRegression
from ..stochastic_process import (
    FrozenNonHomogeneousPoissonProcess,
    RenewalRewardProcess,
)
from ._base import BaseAgeReplacementPolicy, BaseOneCycleAgeReplacementPolicy


class OneCycleAgeReplacementPolicy(BaseOneCycleAgeReplacementPolicy[FrozenAgeReplacementModel, AgeReplacementReward]):
    # noinspection PyUnresolvedReferences
    r"""One-cyle age replacement policy.

    The asset is disposed at a fixed age :math:`a_r` with costs :math:`c_p` or upon failure
    with costs :math:`c_f` if earlier.

    Parameters
    ----------
    lifetime_model : any lifetime distribution or frozen lifetime model
        A lifetime model representing the durations between events.
    cf : float or 1darray
        Costs of failures
    cp : float or 1darray
        Costs of preventive replacements
    discounting_rate : float, default is 0.
        The discounting rate value used in the exponential discounting function
    a0 : float or 1darray, optional
        Current ages of the assets. If it is given, left truncations of ``a0`` will
        be take into account for the first cycle.
    ar : float or 1darray, optional
        Ages of preventive replacements, by default None. If not given, one must call ``optimize`` to set ``ar`` values
        and access to the rest of the object interface.

    Attributes
    ----------
    a0 : float or 1darray, optional
        Current ages of the assets. If it is given, left truncations of ``a0`` will
        be take into account for the first cycle.
    ar
    cf
    cp
    tr


    References
    ----------
    .. [1] Coolen-Schrijner, P., & Coolen, F. P. A. (2006). On optimality
        criteria for age replacement. Proceedings of the Institution of
        Mechanical Engineers, Part O: Journal of Risk and Reliability,
        220(1), 21-29
    """

    lifetime_model: FrozenAgeReplacementModel
    reward: AgeReplacementReward

    def __init__(
        self,
        lifetime_model: LifetimeDistribution | FrozenLifetimeRegression | FrozenLeftTruncatedModel,
        cf: float | NDArray[np.float64],
        cp: float | NDArray[np.float64],
        discounting_rate: float = 0.0,
        period_before_discounting: float = 1.0,
        a0: Optional[float | NDArray[np.float64]] = None,
        ar: Optional[float | NDArray[np.float64]] = None,
    ) -> None:
        self.a0 = np.float64(a0) if isinstance(a0, (float, int)) else a0  # None, arr () or (m,) or (m, 1)
        self._ar = ar if ar is not None else np.nan
        if a0 is not None:
            lifetime_model: FrozenAgeReplacementModel = AgeReplacementModel(LeftTruncatedModel(lifetime_model)).freeze(
                self._ar, a0
            )
        else:
            lifetime_model: FrozenAgeReplacementModel = AgeReplacementModel(lifetime_model).freeze(self._ar)
        reward = AgeReplacementReward(cf, cp, ar)
        super().__init__(lifetime_model, reward, discounting_rate, period_before_discounting)

    @property
    def cf(self) -> NDArray[np.float64]:
        """
        Cost of failures

        Returns
        -------
        ndarray
        """
        return self.reward.cf

    @property
    def cp(self) -> NDArray[np.float64]:
        """
        Cost of preventive replacements

        Returns
        -------
        ndarray
        """
        return self.reward.cp

    @property
    def ar(self) -> float | NDArray[np.float64]:
        """
        Ages of the preventive replacements

        Returns
        -------
        ndarray
        """
        return np.squeeze(self._ar)

    @ar.setter
    def ar(self, value: float | NDArray[np.float64]) -> None:
        self._ar = value
        if self.a0 is None:
            a0 = np.zeros_like(value)
        else:
            a0 = self.a0
        time_before_replacement = np.maximum(value - a0, 0)
        self.lifetime_model.ar = time_before_replacement
        self.reward.ar = time_before_replacement

    @property
    def tr(self) -> float | NDArray[np.float64]:
        """
        Time before the replacement

        Returns
        -------
        ndarray
        """
        return np.squeeze(self.lifetime_model.ar)

    @override
    def expected_total_cost(
        self,
        tf: float,
        nb_steps: int,
    ) -> tuple[NDArray[np.float64], NDArray[np.float64]]:
        if np.any(np.isnan(self.ar)):
            raise ValueError
        return super().expected_total_cost(tf, nb_steps)  # (nb_steps,), (nb_steps,) or (nb_steps,), (m, nb_steps)

    @override
    def asymptotic_expected_total_cost(self) -> NDArray[np.float64]:
        if np.any(np.isnan(self.ar)):
            raise ValueError
        return super().asymptotic_expected_total_cost()  # () or (m, nb_steps)

    @override
    def expected_equivalent_annual_cost(
        self, tf: float, nb_steps: int
    ) -> tuple[NDArray[np.float64], NDArray[np.float64]]:
        if np.any(np.isnan(self.ar)):
            raise ValueError
        # because b = np.minimum(ar, b) in ls_integrate, b can be lower than a depending on period before discounting
        if np.any(self.period_before_discounting >= self.ar):
            raise ValueError("The period before discounting must be lower than ar values")

        if np.any(self.tr == 0):
            warnings.warn(
                "Some assets has already been replaced for the first cycle (where tr is 0). For these assets, consider adjusting ar values to be greater than a0"
            )

        ar = self.ar.copy()
        #  change ar temporarly to enable computation of eeac (if not, AgeReplacementModel.ls_integrate bounds will be problematic)
        self.ar = np.where(self.tr == 0, np.inf, self.ar)
        timeline, eeac = super().expected_equivalent_annual_cost(tf, nb_steps)
        if eeac.ndim == 2:  # more than one asset
            eeac[np.where(np.atleast_1d(self.ar) == np.inf)[0], :] = np.nan
        if eeac.ndim == 1 and self.ar == np.inf:
            eeac.fill(np.nan)
        self.ar = ar
        return timeline, eeac  # (nb_steps,), (nb_steps,) or (nb_steps,), (m, nb_steps)

    @override
    def asymptotic_expected_equivalent_annual_cost(self) -> NDArray[np.float64]:
        if np.any(np.isnan(self.ar)):
            raise ValueError
        # because b = np.minimum(ar, b) in ls_integrate, b can be lower than a depending on period before discounting
        if np.any(self.period_before_discounting >= self.ar):
            raise ValueError("The period before discounting must be lower than ar values")

        if np.any(self.tr == 0):
            warnings.warn(
                "Some assets has already been replaced for the first cycle (where tr is 0). For these assets, consider adjusting ar values to be greater than a0"
            )

        ar = self.ar.copy()
        #  change ar temporarly to enable computation of eeac (if not, AgeReplacementModel.ls_integrate bounds are not in right order)
        self.ar = np.where(self.tr == 0, np.inf, self.ar)
        asymptotic_eeac = super().asymptotic_expected_equivalent_annual_cost()

        if asymptotic_eeac.ndim == 1:  # more than one asset
            asymptotic_eeac[np.where(self.ar == np.inf)[0]] = np.nan
        if asymptotic_eeac.ndim == 0 and self.ar == np.inf:
            asymptotic_eeac = np.nan
        self.ar = ar
        return asymptotic_eeac  # () or (m, nb_steps)

    def optimize(
        self,
    ) -> OneCycleAgeReplacementPolicy:
        """
        Computes the optimal age(s) of replacement and updates the internal ``ar`` value(s) and, optionally ``ar1``.

        Returns
        -------
        Self
             Same instance with optimized ``ar`` (optionnaly ``ar1``).
        """

        x0 = np.minimum(self.cp / (self.cf - self.cp), 1)  # ()
        #  TODO : LeftTruncatedModel.hf (ne pas utiliser LeftTruncatedModel),
        #  retourner ar parfois < a0 (ou ar = 0), mais garder LeftTruncated pour la projection des conséquences
        if self.a0 is not None:
            # FrozenAgeReplacementModel(LeftTruncatedModel(...))
            hf = self.lifetime_model.unfrozen_model.baseline.baseline.hf
        else:
            # FrozenAgeReplacementModel(...)
            hf = self.lifetime_model.unfrozen_model.baseline.hf
        x0 = np.broadcast_to(x0, hf(x0).shape).copy()  # () or (m, 1)

        def eq(a):  # () or (m, 1)
            return (
                self.discounting.factor(a)
                / self.discounting.annuity_factor(a)
                * ((self.cf - self.cp) * hf(a) - self.cp / self.discounting.annuity_factor(a))
            )
            # return ((self.cf - self.cp) / self.cp) * hf(a) * self.discounting.annuity_factor(a)

        ar = np.squeeze(newton(eq, x0))  # () or (m,)

        if self.a0 is not None and self.a0.ndim > 0:
            ar = np.broadcast_to(ar, self.a0.shape).copy()

        self.ar = ar  # setter is called

        return self


class AgeReplacementPolicy(BaseAgeReplacementPolicy[FrozenAgeReplacementModel, AgeReplacementReward]):
    # noinspection PyUnresolvedReferences
    r"""Age replacement policy.

    Behind the scene, a renewal reward stochastic process is used where assets are replaced at a fixed age :math:`a_r`
    with costs :math:`c_p` or upon failure with costs :math:`c_f` if earlier [1]_.

    Parameters
    ----------
    lifetime_model : any lifetime distribution or frozen lifetime model
        A lifetime model representing the durations between events.
    cf : float or 1darray
        Costs of failures
    cp : float or 1darray
        Costs of preventive replacements
    discounting_rate : float, default is 0.
        The discounting rate value used in the exponential discounting function
    a0 : float or 1darray, optional
        Current ages of the assets. If it is given, left truncations of ``a0`` will
        be take into account for the first cycle.
    ar : float or 1darray, optional
        Ages of preventive replacements, by default None. If not given, one must call ``optimize`` to set ``ar`` values
        and access to the rest of the object interface.
    ar1 : float, 2D array, optional
        Ages of the first preventive replacements, by default None. If not given, one must call ``optimize`` to set ``ar`` values
        and access to the rest of the object interface.

    Attributes
    ----------
    a0 : float or 1darray, optional
        Current ages of the assets. If it is given, left truncations of ``a0`` will
        be take into account for the first cycle.
    ar
    first_cycle_tr
    cf
    cp

    References
    ----------
    .. [1] Mazzuchi, T. A., Van Noortwijk, J. M., & Kallen, M. J. (2007).
        Maintenance optimization. Encyclopedia of Statistics in Quality and
        Reliability, 1000-1008.
    """

    stochastic_process: RenewalRewardProcess[FrozenAgeReplacementModel, AgeReplacementReward]

    def __init__(
        self,
        lifetime_model: LifetimeDistribution | FrozenLifetimeRegression,
        cf: float | NDArray[np.float64],
        cp: float | NDArray[np.float64],
        discounting_rate: float = 0.0,
        a0: Optional[float | NDArray[np.float64]] = None,
        ar: Optional[float | NDArray[np.float64]] = None,
        ar1: Optional[float | NDArray[np.float64]] = None,
    ) -> None:

        self.a0 = None
        first_lifetime_model: Optional[FrozenAgeReplacementModel] = None
        if a0 is not None:
            self.a0 = np.float64(a0) if isinstance(a0, (float, int)) else a0  # None, arr () or (m,) or (m, 1)
            first_lifetime_model: FrozenAgeReplacementModel = AgeReplacementModel(LeftTruncatedModel(lifetime_model)).freeze(ar1 if ar1 is not None else np.nan, a0)
        lifetime_model: FrozenAgeReplacementModel = AgeReplacementModel(lifetime_model).freeze(
            ar if ar is not None else np.nan
        )
        reward = AgeReplacementReward(cf, cp, ar if ar is not None else np.nan)

        stochastic_process = RenewalRewardProcess(
            lifetime_model,
            reward,
            discounting_rate,
            first_lifetime_model=first_lifetime_model,
        )
        super().__init__(stochastic_process)

    @property
    def ar(self) -> float | NDArray[np.float64]:
        """
        Ages of the preventive replacements

        Returns
        -------
        ndarray
        """
        return np.squeeze(self.stochastic_process.lifetime_model.ar)  # may be (m, 1) so squeeze it

    @ar.setter
    def ar(self, value: float | NDArray[np.float64]) -> None:

        if self.a0 is not None:
            value = np.broadcast_to(value, self.a0.shape).copy()

        self.stochastic_process.lifetime_model.ar = value
        self.stochastic_process.reward.ar = value

        if self.a0 is not None:
            first_cycle_tr = np.maximum(value - self.a0, 0)
            self.stochastic_process.first_lifetime_model.ar = first_cycle_tr
            self.stochastic_process.first_reward.ar = first_cycle_tr

    @property
    def first_cycle_tr(self) -> float | NDArray[np.float64]:
        """
        Time before the first replacement

        Returns
        -------
        ndarray
        """
        if self.a0 is not None:
            return np.squeeze(self.stochastic_process.first_lifetime_model.ar)
        return self.ar

    @property
    def cf(self) -> float | NDArray[np.float64]:
        """
        Cost of failures

        Returns
        -------
        ndarray
        """
        return self.stochastic_process.reward.cf

    @cf.setter
    def cf(self, value: float | NDArray[np.float64]) -> None:
        self.stochastic_process.reward.cf = value
        self.stochastic_process.first_reward.cf = value

    @property
    def cp(self) -> float | NDArray[np.float64]:
        """
        Cost of preventive replacements

        Returns
        -------
        ndarray
        """
        return self.stochastic_process.reward.cp

    @cp.setter
    def cp(self, value: float | NDArray[np.float64]) -> None:
        self.stochastic_process.reward.cp = value
        self.stochastic_process.first_reward.cp = value

    @override
    def expected_total_cost(
        self,
        tf: float,
        nb_steps: int,
    ) -> tuple[NDArray[np.float64], NDArray[np.float64]]:
        if np.any(np.isnan(self.ar)):
            raise ValueError
        return self.stochastic_process.expected_total_reward(
            tf, nb_steps
        )  # (nb_steps,), (nb_steps,) or (nb_steps,), (m, nb_steps)

    @override
    def asymptotic_expected_total_cost(self) -> NDArray[np.float64]:
        if np.any(np.isnan(self.ar)):
            raise ValueError
        return self.stochastic_process.asymptotic_expected_total_reward()  # () or (m,)

    @override
    def expected_equivalent_annual_cost(
        self,
        tf: float,
        nb_steps: int,
    ) -> tuple[NDArray[np.float64], NDArray[np.float64]]:
        if np.any(np.isnan(self.ar)):
            raise ValueError
        timeline, eeac = self.stochastic_process.expected_equivalent_annual_worth(tf, nb_steps)
        if np.any(self.first_cycle_tr == 0):
            warnings.warn(
                "Some assets has already been replaced for the first cycle (where first_cycle_tr is 0). For these assets, consider adjusting ar values to be greater than a0"
            )
        if eeac.ndim == 2:
            eeac[np.where(np.atleast_1d(self.first_cycle_tr) == 0)] = np.nan
        if eeac.ndim == 1 and self.ar == np.inf:
            eeac.fill(np.nan)
        return timeline, eeac  # (nb_steps,), (nb_steps,) or (nb_steps,), (m, nb_steps)

    @override
    def asymptotic_expected_equivalent_annual_cost(self) -> NDArray[np.float64]:
        if np.any(np.isnan(self.ar)):
            raise ValueError
        asymptotic_eeac = self.stochastic_process.asymptotic_expected_equivalent_annual_worth()
        if np.any(self.first_cycle_tr == 0):
            warnings.warn(
                "Some assets has already been replaced for the first cycle (where first_cycle_tr is 0). For these assets, consider adjusting ar values to be greater than a0"
            )
        if asymptotic_eeac.ndim == 1:
            asymptotic_eeac[np.where(np.atleast_1d(self.first_cycle_tr) == 0)] = np.nan
        if asymptotic_eeac.ndim == 0 and self.first_cycle_tr == 0:
            asymptotic_eeac = np.nan
        return asymptotic_eeac  # () or (m,)

    def annual_number_of_replacements(
        self,
        nb_years: int,
        upon_failure: bool = False,
        total: bool = True,
    ):
        """
        The expected number of annual replacements.

        Parameters
        ----------
        nb_years : int
            The number of years on which the annual number of replacements are projected
        upon_failure : bool, default is False
            If True, it also returns the annual number of replacements due to unexpected failures
        total : bool, default is True
            If True, the given numbers of replacements are the sum of all replacements without distinction between assets
        """
        if np.any(np.isnan(self.ar)):
            raise ValueError

        copied_policy = copy.deepcopy(self)
        copied_policy.cf = 1.0
        copied_policy.cp = 1.0
        copied_policy.discounting_rate = 0.0
        timeline, total_cost = copied_policy.expected_total_cost(nb_years, nb_years + 1)  # equiv to np.arange
        if total:
            mt = np.sum(np.atleast_2d(total_cost), axis=0)
        else:
            mt = total_cost
        nb_replacements = np.diff(mt)
        if upon_failure:
            copied_policy.cp = 0.0
            _, total_cost = copied_policy.expected_total_cost(nb_years, nb_years + 1)  # equiv to np.arange
            if total:
                mf = np.sum(np.atleast_2d(total_cost), axis=0)
            else:
                mf = total_cost
            nb_failures = np.diff(mf)
            return timeline[1:], nb_replacements, nb_failures
        return timeline[1:], nb_replacements

    def optimize(
        self,
    ) -> AgeReplacementPolicy:
        """
        Computes the optimal age(s) of replacement and updates the internal ``ar`` value(s) and,t optionally ``ar1``.

        Returns
        -------
        Self
             Same instance with optimized ``ar`` (optionnaly ``ar1``).
        """
        x0 = np.minimum(self.cp / (self.cf - self.cp), 1)  # ()
        discounting = ExponentialDiscounting(self.stochastic_process.discounting_rate)

        sf = self.stochastic_process.lifetime_model.unfrozen_model.baseline.sf
        pdf = self.stochastic_process.lifetime_model.unfrozen_model.baseline.pdf
        hf = self.stochastic_process.lifetime_model.unfrozen_model.baseline.hf
        x0 = np.broadcast_to(x0, hf(x0).shape).copy()  # () or (m, 1)

        def eq(a):  # () or (m, 1)
            f = legendre_quadrature(
                lambda x: discounting.factor(x) * sf(x),
                0,
                a,
            )
            g = legendre_quadrature(
                lambda x: discounting.factor(x) * pdf(x),
                0,
                a,
            )
            return discounting.factor(a) * ((self.cf - self.cp) * (hf(a) * f - g) - self.cp) / f**2

        self.ar = np.squeeze(newton(eq, x0))  # () or (m,) setter is called

        return self


class NonHomogeneousPoissonAgeReplacementPolicy:
    # noinspection PyUnresolvedReferences
    r"""Age replacement policy for non-Homogeneous Poisson process.

    Parameters
    ----------
    process : non-Homogeneous Poisson process
        The underlying process. If the process expects covars, it must be frozen before.
    cr : float or 1darray
        The cost of repair.
    cp : float or 1darray
        The cost of failure.
    discounting_rate : float, default is 0.
        The discounting rate value used in the exponential discounting function
    ar : float or 1darray, optional
        Ages of preventive replacements, by default None. If not given, one must call ``optimize`` to set ``ar`` values
        and access to the rest of the object interface.

    Attributes
    ----------
    ar
    cr
    cp
    discounting_rate
    """

    def __init__(
        self,
        process: FrozenNonHomogeneousPoissonProcess,
        cr: float | NDArray[np.float64],
        cp: float | NDArray[np.float64],
        discounting_rate: float = 0.0,
        ar: Optional[float | NDArray[np.float64]] = None,
    ) -> None:

        self.ar = ar if ar is not None else np.nan
        self.cost = cost(cr=cr, cp=cp)
        self.discounting = ExponentialDiscounting(discounting_rate)
        self.process = process

    @property
    def discounting_rate(self):
        """
        The discounting rate

        Returns
        -------
        float
        """
        return self.discounting.rate

    @property
    def cr(self) -> float | NDArray[np.float64]:
        """
        Cost of repair

        Returns
        -------
        ndarray
        """
        return self.cost["cr"]

    @cr.setter
    def cr(self, value: float | NDArray[np.float64]) -> None:
        self.cost["cr"] = value

    @property
    def cp(self) -> float | NDArray[np.float64]:
        """
        Cost of preventive replacements

        Returns
        -------
        ndarray
        """
        return self.cost["cp"]

    @cp.setter
    def cp(self, value: float | NDArray[np.float64]) -> None:
        self.cost["cp"] = value

    @property
    def ar(self):
        """
        Ages of the preventive replacements

        Returns
        -------
        ndarray
        """
        return np.squeeze(self._ar)

    @ar.setter
    def ar(self, value):
        ar = np.asarray(value, dtype=np.float64)
        shape = () if ar.ndim == 0 else (ar.size, 1)
        self._ar = ar.reshape(shape)

    def asymptotic_expected_equivalent_annual_cost(self) -> NDArray[np.float64]:
        r"""
        The asymtotic expected equivalent annual cost

        .. math::

            \lim_{t\to\infty} \text{EEAC}(t)

        where :math:`\text{EEAC}(t)` is the expected equivalent annual cost at :math:`t`.

        Returns
        -------
        ndarray
            The asymptotic expected equivalent annual cost
        """
        if np.any(np.isnan(self.ar)):
            raise ValueError
        if self.discounting_rate == 0.0:
            asymptotic_eeac = (
                self.cp + self.cr * legendre_quadrature(lambda t: self.process.intensity(t), 0, self._ar)
            ) / self._ar
        else:
            asymptotic_eeac = (
                self.discounting_rate
                * (
                    self.cp * self.discounting.factor(self._ar)
                    + self.cr
                    * legendre_quadrature(lambda t: self.discounting.factor(t) * self.process.intensity(t), 0, self._ar)
                )
                / (1 - self.discounting.factor(self._ar))
            )
        return np.squeeze(asymptotic_eeac)  # () or (m,)

    def optimize(
        self,
    ) -> NDArray[np.float64]:
        """
        Computes the optimal age(s) of replacement and updates the internal ``ar`` value(s).

        Returns
        -------
        Self
             Same instance with optimized ``ar``.
        """

        x0 = self.process.unfreeze().lifetime_model.mean(*self.process.args)
        x0 = np.broadcast_to(
            x0, np.broadcast_shapes(self.process.intensity(x0).shape, self.cp.shape, self.cr.shape)
        ).copy()

        def eq(ar):
            ar = np.atleast_2d(ar)
            if self.discounting.rate != 0:
                return (
                    (1 - self.discounting.factor(ar)) / self.discounting_rate * self.process.intensity(ar)
                    - legendre_quadrature(
                        lambda t: self.discounting.factor(t) * self.process.intensity(t),
                        np.array(0.0),
                        ar,
                    )
                    - self.cp / self.cr
                )
            return ar * self.process.intensity(ar) - self.process.cumulative_intensity(ar) - self.cp / self.cr

        self.ar = np.squeeze(newton(eq, x0))
        return self


from ._docstring import (
    ASYMPTOTIC_EEAC_DOCSTRING,
    ASYMPTOTIC_ETC_DOCSTRING,
    EEAC_DOCSTRING,
    ETC_DOCSTRING,
)

WARNING = r"""

.. warning::

    This method requires the ``ar`` attribute to be set either at initialization
    or with the ``optimize`` method.
"""

OneCycleAgeReplacementPolicy.expected_total_cost.__doc__ = ETC_DOCSTRING + WARNING
OneCycleAgeReplacementPolicy.expected_equivalent_annual_cost.__doc__ = EEAC_DOCSTRING + WARNING
OneCycleAgeReplacementPolicy.asymptotic_expected_total_cost.__doc__ = ASYMPTOTIC_ETC_DOCSTRING + WARNING
OneCycleAgeReplacementPolicy.asymptotic_expected_equivalent_annual_cost.__doc__ = ASYMPTOTIC_EEAC_DOCSTRING + WARNING



import numpy as np

from ._base import FrozenParametricLifetimeModel, ParametricLifetimeModel


def reshape_ar_or_a0(name: str, value):
    value = np.asarray(value)  # in shape : (), (m,) or (m, 1)
    if value.ndim > 2 or (value.ndim == 2 and value.shape[-1] != 1):
        raise ValueError(f"Incorrect {name} shape. Got {value.shape}. Expected (), (m,) or (m, 1)")
    if value.ndim == 1:
        value = value.reshape(-1, 1)
    return value  # out shape: () or (m, 1)


class AgeReplacementModel(ParametricLifetimeModel):
    # noinspection PyUnresolvedReferences
    r"""
    Age replacement model.

    Lifetime model where the assets are replaced at age :math:`a_r`. This is equivalent to the model of :math:`\min(X,a_r)` where
    :math:`X` is a baseline lifetime and :math:`a_r` is the age of replacement.

    Parameters
    ----------
    baseline : any parametric lifetime model (frozen lifetime model works)
        The base lifetime model without conditional probabilities

    Attributes
    ----------
    baseline
    nb_params
    params
    params_names
    plot
    """

    def __init__(self, baseline):
        super().__init__()
        self.baseline = baseline

    def sf(self, time, ar, *args):
        """
        The survival function.

        Parameters
        ----------
        time : float or np.ndarray
            Elapsed time value(s) at which to compute the function.
            If ndarray, allowed shapes are ``()``, ``(n_values,)`` or ``(n_assets, n_values)``.
        ar : float or np.ndarray
            Age of replacement values. If ndarray, shape can only be (m,)
            as only one age of replacement per asset can be given
        *args : float or np.ndarray
            Additional arguments needed by the model.

        Returns
        -------
        np.float64 or np.ndarray
            Function values at each given time(s).
        """
        ar = reshape_ar_or_a0("ar", ar)
        return np.where(time < ar, self.baseline.sf(time, *args), 0.0)

    def hf(self, time, ar, *args):
        """
        The hazard function.

        Parameters
        ----------
        time : float or np.ndarray
            Elapsed time value(s) at which to compute the function.
            If ndarray, allowed shapes are ``()``, ``(n_values,)`` or ``(n_assets, n_values)``.
        ar : float or np.ndarray
            Age of replacement values. If ndarray, shape can only be (m,)
            as only one age of replacement per asset can be given
        *args : float or np.ndarray
            Additional arguments needed by the model.

        Returns
        -------
        np.float64 or np.ndarray
            Function values at each given time(s).
        """
        ar = reshape_ar_or_a0("ar", ar)
        return np.where(time < ar, self.baseline.hf(time, *args), 0.0)

    def cdf(self, time, ar, *args):
        """
        The cumulative density function.

        Parameters
        ----------
        time : float or np.ndarray
            Elapsed time value(s) at which to compute the function.
            If ndarray, allowed shapes are ``()``, ``(n_values,)`` or ``(n_assets, n_values)``.
        ar : float or np.ndarray
            Age of replacement values. If ndarray, shape can only be (m,)
            as only one age of replacement per asset can be given
        *args : float or np.ndarray
            Additional arguments needed by the model.

        Returns
        -------
        np.float64 or np.ndarray
            Function values at each given time(s).
        """
        ar = reshape_ar_or_a0("ar", ar)
        return super().cdf(time, *(ar, *args))

    def chf(self, time, ar, *args):
        """
        The cumulative hazard function.

        Parameters
        ----------
        time : float or np.ndarray
            Elapsed time value(s) at which to compute the function.
            If ndarray, allowed shapes are ``()``, ``(n_values,)`` or ``(n_assets, n_values)``.
        ar : float or np.ndarray
            Age of replacement values. If ndarray, shape can only be (m,)
            as only one age of replacement per asset can be given
        *args : float or np.ndarray
            Additional arguments needed by the model.

        Returns
        -------
        np.float64 or np.ndarray
            Function values at each given time(s).
        """
        ar = reshape_ar_or_a0("ar", ar)
        return np.where(time < ar, self.baseline.chf(time, *args), 0.0)

    def isf(self, probability, ar, *args):
        """
        The inverse of the survival function.

        Parameters
        ----------
        probability : float or np.ndarray
            Probability value(s) at which to compute the function.
            If ndarray, allowed shapes are ``()``, ``(n,)`` or ``(m, n)``.
        ar : float or np.ndarray
            Age of replacement values. If ndarray, shape can only be (m,)
            as only one age of replacement per asset can be given
        *args : float or np.ndarray
            Additional arguments needed by the model.

        Returns
        -------
        np.float64 or np.ndarray
            Function values at each given probability value(s).
        """
        ar = reshape_ar_or_a0("ar", ar)
        return np.minimum(self.baseline.isf(probability, *args), ar)

    def ichf(
        self,
        cumulative_hazard_rate,
        ar,
        *args,
    ):
        """
        Inverse cumulative hazard function.

        Parameters
        ----------
        cumulative_hazard_rate : float or np.ndarray
            Cumulative hazard rate value(s) at which to compute the function.
            If ndarray, allowed shapes are ``()``, ``(n,)`` or ``(m, n)``.

        ar : float or np.ndarray
            Age of replacement values. If ndarray, shape can only be (m,)
            as only one age of replacement per asset can be given
        *args : float or np.ndarray
            Additional arguments needed by the model.

        Returns
        -------
        np.float64 or np.ndarray
            Function values at each given cumulative hazard rate(s).
        """
        ar = reshape_ar_or_a0("ar", ar)
        return np.minimum(self.baseline.ichf(cumulative_hazard_rate, *args), ar)

    def pdf(self, time, ar, *args):
        """
        The probability density function.

        Parameters
        ----------
        time : float or np.ndarray
            Elapsed time value(s) at which to compute the function.
            If ndarray, allowed shapes are ``()``, ``(n_values,)`` or ``(n_assets, n_values)``.
        ar : float or np.ndarray
            Age of replacement values. If ndarray, shape can only be (m,)
            as only one age of replacement per asset can be given
        *args : float or np.ndarray
            Additional arguments needed by the model.

        Returns
        -------
        np.float64 or np.ndarray
            Function values at each given time(s).
        """
        ar = reshape_ar_or_a0("ar", ar)
        return np.where(time < ar, self.baseline.pdf(time, *args), 0)

    def mrl(self, time, ar, *args):
        """
        The mean residual life function.

        Parameters
        ----------
        time : float or np.ndarray
            Elapsed time value(s) at which to compute the function.
            If ndarray, allowed shapes are ``()``, ``(n_values,)`` or ``(n_assets, n_values)``.
        ar : float or np.ndarray
            Age of replacement values. If ndarray, shape can only be (m,)
            as only one age of replacement per asset can be given
        *args : float or np.ndarray
            Additional arguments needed by the model.

        Returns
        -------
        np.float64 or np.ndarray
            Function values at each given time(s).
        """
        ar = reshape_ar_or_a0("ar", ar)
        ub = np.array(np.inf)
        # ar.shape == (m, 1)
        mask = time >= ar  # (m, 1) or (m, n)
        if np.any(mask):
            time, ub = np.broadcast_arrays(time, ub)
            time = np.ma.MaskedArray(time, mask)  # (m, 1) or (m, n)
            ub = np.ma.MaskedArray(ub, mask)  # (m, 1) or (m, n)
        mu = self.ls_integrate(lambda x: x - time, time, ub, ar, *args, deg=10) / self.sf(
            time, ar, *args
        )  # () or (n,) or (m, n)
        np.ma.filled(mu, 0)
        return np.ma.getdata(mu)

    def ppf(self, probability, ar, *args):
        """
        The percent point function.

        Parameters
        ----------
        probability : float or np.ndarray
            Probability value(s) at which to compute the function.
            If ndarray, allowed shapes are ``()``, ``(n,)`` or ``(m, n)``.
        ar : float or np.ndarray
            Age of replacement values. If ndarray, shape can only be (m,)
            as only one age of replacement per asset can be given
        *args : float or np.ndarray
            Additional arguments needed by the model.

        Returns
        -------
        np.float64 or np.ndarray
            Function values at each given probability value(s).
        """
        ar = reshape_ar_or_a0("ar", ar)
        return self.isf(1 - probability, ar, *args)

    # def cdf(self, time, ar, *args):
    #     ar = reshape_ar_or_a0("ar", ar)
    #     return np.where(time < ar, self.baseline.cdf(time, *args), 1.0)

    def median(self, ar, *args):
        """
        The median.

        Parameters
        ----------
        ar : float or np.ndarray
            Age of replacement values. If ndarray, shape can only be (m,)
            as only one age of replacement per asset can be given
        *args : float or np.ndarray
            Additional arguments needed by the model.

        Returns
        -------
        np.float64 or np.ndarray
        """
        ar = reshape_ar_or_a0("ar", ar)
        return self.ppf(np.array(0.5), ar, *args)

    def rvs(self, size, ar, *args, nb_assets=None, return_event=False, return_entry=False, seed=None):
        """
        Random variable sampling.

        Parameters
        ----------
        size : int
            Size of the generated sample.
        ar : float or np.ndarray
            Age of replacement values. If ndarray, shape can only be (m,)
            as only one age of replacement per asset can be given
        *args : float or np.ndarray
            Additional arguments needed by the model.
        nb_assets : int, optional
            If nb_assets is not None, 2d arrays of samples are generated.
        return_event : bool, default is False
            If True, returns event indicators along with the sample time values.
        return_entry : bool, default is False
            If True, returns corresponding entry values of the sample time values.
        seed : optional int, default is None
            Random seed used to fix random sampling.

        Returns
        -------
        float, ndarray or tuple of float or ndarray
            The sample values. If either ``return_event`` or ``return_entry`` is True, returns a tuple containing
            the time values followed by event values, entry values or both.

        Notes
        -----
        If ``return_entry`` is true, returned time values are not residual time. Otherwise, the times are residuals
        """
        ar = reshape_ar_or_a0("ar", ar)
        baseline_rvs = self.baseline.rvs(
            size, *args, nb_assets=nb_assets, return_event=return_event, return_entry=return_entry, seed=seed
        )
        time = baseline_rvs[0] if isinstance(baseline_rvs, tuple) else baseline_rvs
        time = np.minimum(time, ar)  # it may change time shape by broadcasting
        if not return_event and not return_entry:
            return time
        elif return_event and not return_entry:
            event = np.broadcast_to(baseline_rvs[1], time.shape).copy()
            event = np.where(time != ar, event, ~event)
            return time, event
        elif not return_event and return_entry:
            entry = np.broadcast_to(baseline_rvs[1], time.shape).copy()
            return time, entry
        else:
            event, entry = baseline_rvs[1:]
            event = np.broadcast_to(event, time.shape).copy()
            entry = np.broadcast_to(entry, time.shape).copy()
            event = np.where(time != ar, event, ~event)
            return time, event, entry

    def ls_integrate(self, func, a, b, ar, *args, deg=10):
        """
        Lebesgue-Stieltjes integration.

        Parameters
        ----------
        func : callable (in : 1 ndarray , out : 1 ndarray)
            The callable must have only one ndarray object as argument and one ndarray object as output
        a : ndarray (maximum number of dimension is 2)
            Lower bound(s) of integration.
        b : ndarray (maximum number of dimension is 2)
            Upper bound(s) of integration. If lower bound(s) is infinite, use np.inf as value.)
        ar : float or np.ndarray
            Age of replacement values. If ndarray, shape can only be (m,)
            as only one age of replacement per asset can be given
        *args : float or np.ndarray
            Additional arguments needed by the model.
        deg : int, default 10
            Degree of the polynomials interpolation

        Returns
        -------
        np.ndarray
            Lebesgue-Stieltjes integral of func from `a` to `b`.
        """
        ar = reshape_ar_or_a0("ar", ar)
        b = np.minimum(ar, b)
        integration = self.baseline.ls_integrate(func, a, b, *args, deg=deg)
        return integration + np.where(b == ar, func(ar) * self.baseline.sf(ar, *args), 0)

    def moment(self, n, ar, *args):
        """
        n-th order moment

        Parameters
        ----------
        n : order of the moment, at least 1.
        ar : float or np.ndarray
            Age of replacement values. If ndarray, shape can only be (m,)
            as only one age of replacement per asset can be given
        *args : float or np.ndarray
            Additional arguments needed by the model.

        Returns
        -------
        np.float64 or np.ndarray
        """
        ar = reshape_ar_or_a0("ar", ar)
        return self.ls_integrate(
            lambda x: x**n,
            0,
            np.inf,
            ar,
            *args,
            deg=100,
        )

    def mean(self, ar, *args):
        """
        The mean.

        Parameters
        ----------
        ar : float or np.ndarray
            Age of replacement values. If ndarray, shape can only be (m,)
            as only one age of replacement per asset can be given
        *args : float or np.ndarray
            Additional arguments needed by the model.

        Returns
        -------
        np.float64 or np.ndarray
        """
        ar = reshape_ar_or_a0("ar", ar)
        return self.moment(1, ar, *args)

    def var(self, ar, *args):
        """
        The variance.

        Parameters
        ----------
        ar : float or np.ndarray
            Age of replacement values. If ndarray, shape can only be (m,)
            as only one age of replacement per asset can be given
        *args : float or np.ndarray
            Additional arguments needed by the model.

        Returns
        -------
        np.float64 or np.ndarray
        """
        ar = reshape_ar_or_a0("ar", ar)
        return self.moment(2, ar, *args) - self.moment(1, ar, *args) ** 2

    def freeze(self, ar, *args):
        """
        Freeze age replacement values and other arguments into the object data.

        Parameters
        ----------
        ar : float or np.ndarray
            Age of replacement values. If ndarray, shape can only be (m,)
            as only one age of replacement per asset can be given
        *args : float or np.ndarray
            Additional arguments needed by the model.

        Returns
        -------
        FrozenAgeReplacementModel
        """
        ar = reshape_ar_or_a0("ar", ar)
        return FrozenAgeReplacementModel(self, ar, *args)


class LeftTruncatedModel(ParametricLifetimeModel):
    # noinspection PyUnresolvedReferences
    r"""Left truncated model.

    Lifetime model where the assets have already reached the age :math:`a_0`.

    Parameters
    ----------
    baseline : any parametric lifetime model (frozen lifetime model works)
        The base lifetime model without conditional probabilities
    nb_params
    params
    params_names
    plot

    Attributes
    ----------
    baseline
    nb_params
    params
    params_names
    plot
    """

    def __init__(self, baseline):
        super().__init__()
        self.baseline = baseline

    def sf(self, time, a0, *args):
        """
        The survival function.

        Parameters
        ----------
        time : float or np.ndarray
            Elapsed time value(s) at which to compute the function.
            If ndarray, allowed shapes are ``()``, ``(n_values,)`` or ``(n_assets, n_values)``.
        a0 : float or np.ndarray
            Conditional age values. It represents ages reached by assets. If ndarray, shape can only be (m,)
            as only one age per asset can be given
        *args : float or np.ndarray
            Additional arguments needed by the model.

        Returns
        -------
        np.float64 or np.ndarray
            Function values at each given time(s).
        """
        a0 = reshape_ar_or_a0("a0", a0)
        return super().sf(time, a0, *args)

    def pdf(self, time, a0, *args):
        """
        The probability density function.

        Parameters
        ----------
        time : float or np.ndarray
            Elapsed time value(s) at which to compute the function.
            If ndarray, allowed shapes are ``()``, ``(n_values,)`` or ``(n_assets, n_values)``.
        a0 : float or np.ndarray
            Conditional age values. It represents ages reached by assets. If ndarray, shape can only be (m,)
            as only one age per asset can be given
        *args : float or np.ndarray
            Additional arguments needed by the model.

        Returns
        -------
        np.float64 or np.ndarray
            Function values at each given time(s).
        """
        a0 = reshape_ar_or_a0("a0", a0)
        return super().pdf(time, a0, *args)

    def isf(self, probability, a0, *args):
        """
        The inverse of the survival function.

        Parameters
        ----------
        probability : float or np.ndarray
            Probability value(s) at which to compute the function.
            If ndarray, allowed shapes are ``()``, ``(n,)`` or ``(m, n)``.
        a0 : float or np.ndarray
            Conditional age values. It represents ages reached by assets. If ndarray, shape can only be (m,)
            as only one age per asset can be given
        *args : float or np.ndarray
            Additional arguments needed by the model.

        Returns
        -------
        np.float64 or np.ndarray
            Function values at each given probability value(s).
        """
        cumulative_hazard_rate = -np.log(probability + 1e-6)  # avoid division by zero
        a0 = reshape_ar_or_a0("a0", a0)
        return self.ichf(cumulative_hazard_rate, a0, *args)

    def chf(self, time, a0, *args):
        """
        The cumulative hazard function.

        Parameters
        ----------
        time : float or np.ndarray
            Elapsed time value(s) at which to compute the function.
            If ndarray, allowed shapes are ``()``, ``(n_values,)`` or ``(n_assets, n_values)``.
        a0 : float or np.ndarray
            Conditional age values. It represents ages reached by assets. If ndarray, shape can only be (m,)
            as only one age per asset can be given
        *args : float or np.ndarray
            Additional arguments needed by the model.

        Returns
        -------
        np.float64 or np.ndarray
            Function values at each given time(s).
        """
        a0 = reshape_ar_or_a0("a0", a0)
        return self.baseline.chf(a0 + time, *args) - self.baseline.chf(a0, *args)

    def cdf(self, time, a0, *args):
        """
        The cumulative density function.

        Parameters
        ----------
        time : float or np.ndarray
            Elapsed time value(s) at which to compute the function.
            If ndarray, allowed shapes are ``()``, ``(n_values,)`` or ``(n_assets, n_values)``.
        a0 : float or np.ndarray
            Conditional age values. It represents ages reached by assets. If ndarray, shape can only be (m,)
            as only one age per asset can be given
        *args : float or np.ndarray
            Additional arguments needed by the model.

        Returns
        -------
        np.float64 or np.ndarray
            Function values at each given time(s).
        """
        a0 = reshape_ar_or_a0("a0", a0)
        return super().cdf(time, *(a0, *args))

    def hf(self, time, a0, *args):
        """
        The hazard function.

        Parameters
        ----------
        time : float or np.ndarray
            Elapsed time value(s) at which to compute the function.
            If ndarray, allowed shapes are ``()``, ``(n_values,)`` or ``(n_assets, n_values)``.
        a0 : float or np.ndarray
            Conditional age values. It represents ages reached by assets. If ndarray, shape can only be (m,)
            as only one age per asset can be given
        *args : float or np.ndarray
            Additional arguments needed by the model.

        Returns
        -------
        np.float64 or np.ndarray
            Function values at each given time(s).
        """
        a0 = reshape_ar_or_a0("a0", a0)
        return self.baseline.hf(a0 + time, *args)

    def ichf(self, cumulative_hazard_rate, a0, *args):
        """
        Inverse cumulative hazard function.

        Parameters
        ----------
        cumulative_hazard_rate : float or np.ndarray
            Cumulative hazard rate value(s) at which to compute the function.
            If ndarray, allowed shapes are ``()``, ``(n,)`` or ``(m, n)``.

        a0 : float or np.ndarray
            Conditional age values. It represents ages reached by assets. If ndarray, shape can only be (m,)
            as only one age per asset can be given
        *args : float or np.ndarray
            Additional arguments needed by the model.

        Returns
        -------
        np.float64 or np.ndarray
            Function values at each given cumulative hazard rate(s).
        """
        a0 = reshape_ar_or_a0("a0", a0)
        return self.baseline.ichf(cumulative_hazard_rate + self.baseline.chf(a0, *args), *args) - a0

    def rvs(self, size, a0, *args, nb_assets=None, return_event=False, return_entry=False, seed=None):
        """
        Random variable sampling.

        Parameters
        ----------
        size : int
            Size of the generated sample.
        a0 : float or np.ndarray
            Conditional age values. It represents ages reached by assets. If ndarray, shape can only be (m,)
            as only one age per asset can be given
        *args : float or np.ndarray
            Additional arguments needed by the model.
        nb_assets : int, optional
            If nb_assets is not None, 2d arrays of samples are generated.
        return_event : bool, default is False
            If True, returns event indicators along with the sample time values.
        return_entry : bool, default is False
            If True, returns corresponding entry values of the sample time values.
        seed : optional int, default is None
            Random seed used to fix random sampling.

        Returns
        -------
        float, ndarray or tuple of float or ndarray
            The sample values. If either ``return_event`` or ``return_entry`` is True, returns a tuple containing
            the time values followed by event values, entry values or both.
        """
        a0 = reshape_ar_or_a0("a0", a0)
        super_rvs = super().rvs(
            size, *(a0, *args), nb_assets=nb_assets, return_event=return_event, return_entry=return_entry, seed=seed
        )
        if not return_event and return_entry:
            time, entry = super_rvs
            entry = np.broadcast_to(a0, entry.shape).copy()
            time = time + a0  #  not residual age
            return time, entry
        elif return_event and return_entry:
            time, event, entry = super_rvs
            entry = np.broadcast_to(a0, entry.shape).copy()
            time = time + a0  #  not residual age
            return time, event, entry
        else:
            return super_rvs

    def ls_integrate(self, func, a, b, a0, *args, deg=10):
        """
        Lebesgue-Stieltjes integration.

        Parameters
        ----------
        func : callable (in : 1 ndarray , out : 1 ndarray)
            The callable must have only one ndarray object as argument and one ndarray object as output
        a : ndarray (maximum number of dimension is 2)
            Lower bound(s) of integration.
        b : ndarray (maximum number of dimension is 2)
            Upper bound(s) of integration. If lower bound(s) is infinite, use np.inf as value.)
        a0 : float or np.ndarray
            Conditional age values. It represents ages reached by assets. If ndarray, shape can only be (m,)
            as only one age per asset can be given
        *args : float or np.ndarray
            Additional arguments needed by the model.
        deg : int, default 10
            Degree of the polynomials interpolation

        Returns
        -------
        np.ndarray
            Lebesgue-Stieltjes integral of func from `a` to `b`.
        """
        a0 = reshape_ar_or_a0("a0", a0)
        return super().ls_integrate(func, a, b, *(a0, *args), deg=deg)

    def mean(self, a0, *args):
        """
        The mean.

        Parameters
        ----------
        a0 : float or np.ndarray
            Conditional age values. It represents ages reached by assets. If ndarray, shape can only be (m,)
            as only one age per asset can be given
        *args : float or np.ndarray
            Additional arguments needed by the model.

        Returns
        -------
        np.float64 or np.ndarray
        """
        a0 = reshape_ar_or_a0("a0", a0)
        return super().mean(*(a0, *args))

    def median(self, a0, *args):
        """
        The median.

        Parameters
        ----------
        a0 : float or np.ndarray
            Conditional age values. It represents ages reached by assets. If ndarray, shape can only be (m,)
            as only one age per asset can be given
        *args : float or np.ndarray
            Additional arguments needed by the model.

        Returns
        -------
        np.float64 or np.ndarray
        """
        a0 = reshape_ar_or_a0("a0", a0)
        return super().median(*(a0, *args))

    def var(self, a0, *args):
        """
        The variance.

        Parameters
        ----------
        a0 : float or np.ndarray
            Conditional age values. It represents ages reached by assets. If ndarray, shape can only be (m,)
            as only one age per asset can be given
        *args : float or np.ndarray
            Additional arguments needed by the model.

        Returns
        -------
        np.float64 or np.ndarray
        """
        a0 = reshape_ar_or_a0("a0", a0)
        return super().var(*(a0, *args))

    def moment(self, n: int, a0, *args):
        """
        n-th order moment

        Parameters
        ----------
        n : int
            Order of the moment, at least 1
        a0 : float or np.ndarray
            Conditional age values. It represents ages reached by assets. If ndarray, shape can only be (m,)
            as only one age per asset can be given
        *args : float or np.ndarray
            Additional arguments needed by the model.

        Returns
        -------
        np.float64 or np.ndarray
        """
        a0 = reshape_ar_or_a0("a0", a0)
        return super().moment(n, *(a0, *args))

    def mrl(self, time, a0, *args):
        """
        The mean residual life function.

        Parameters
        ----------
        time : float or np.ndarray
            Elapsed time value(s) at which to compute the function.
            If ndarray, allowed shapes are ``()``, ``(n_values,)`` or ``(n_assets, n_values)``.
        a0 : float or np.ndarray
            Conditional age values. It represents ages reached by assets. If ndarray, shape can only be (m,)
            as only one age per asset can be given
        *args : float or np.ndarray
            Additional arguments needed by the model.

        Returns
        -------
        np.float64 or np.ndarray
            Function values at each given time(s).
        """
        a0 = reshape_ar_or_a0("a0", a0)
        return super().mrl(time, *(a0, *args))

    def ppf(self, probability, a0, *args):
        """
        The percent point function.

        Parameters
        ----------
        probability : float or np.ndarray
            Probability value(s) at which to compute the function.
            If ndarray, allowed shapes are ``()``, ``(n,)`` or ``(m, n)``.
        a0 : float or np.ndarray
            Conditional age values. It represents ages reached by assets. If ndarray, shape can only be (m,)
            as only one age per asset can be given
        *args : float or np.ndarray
            Additional arguments needed by the model.

        Returns
        -------
        np.float64 or np.ndarray
            Function values at each given probability value(s).

        Notes
        -----
        The ``ppf`` is the inverse of :py:meth:`~LeftTruncatedModel.cdf`.

        """
        a0 = reshape_ar_or_a0("a0", a0)
        return super().ppf(probability, *(a0, *args))

    def freeze(self, a0, *args):
        """
        Freeze conditional age values and other arguments into the object data.

        Parameters
        ----------
        a0 : float or np.ndarray
            Conditional age values. It represents ages reached by assets. If ndarray, shape can only be (m,)
            as only one age per asset can be given
        *args : float or np.ndarray
            Additional arguments needed by the model.

        Returns
        -------
        FrozenLeftTruncatedModel
        """
        a0 = reshape_ar_or_a0("a0", a0)
        return FrozenLeftTruncatedModel(self, a0, *args)


class FrozenAgeReplacementModel(FrozenParametricLifetimeModel):
    r"""
    Frozen age replacement model.

    Parameters
    ----------
    model : AgeReplacementModel
        Any age replacement model.
    ar : float or np.ndarray
        Age of replacement values to be frozen.
    *args : float or np.ndarray
        Additional arguments needed by the model to be frozen.

    Attributes
    ----------
    unfrozen_model : AgeReplacementModel
        The unfrozen age replacement model.
    args : tuple of float or np.ndarray
        All the frozen arguments given and necessary to compute model functions.
    nb_assets : int
        Number of assets passed in frozen arguments. The data is mainly used to control numpy broadcasting and may not
        interest an user.

    Warnings
    --------
    The recommanded way to instanciate a frozen model is by using``freeze`` factory function.
    """

    def __init__(self, model, ar, *args):
        super().__init__(model, *(ar, *args))

    @property
    def ar(self):
        return self.args[0]

    @ar.setter
    def ar(self, value):
        self.args = (value,) + self.args[1:]


class FrozenLeftTruncatedModel(FrozenParametricLifetimeModel):
    r"""
    Frozen left truncated model.

    Parameters
    ----------
    model : LeftTruncatedModel
        Any left truncated model.
    a0 : float or np.ndarray
        Conditional age values to be frozen.
    *args : float or np.ndarray
        Additional arguments needed by the model to be frozen.


    Attributes
    ----------
    unfrozen_model : LeftTruncatedModel
        The unfrozen left truncated model.
    args : tuple of float or np.ndarray
        All the frozen arguments given and necessary to compute model functions.
    nb_assets : int
        Number of assets passed in frozen arguments. The data is mainly used to control numpy broadcasting and may not
        interest an user.

    Warnings
    --------
    The recommanded way to instanciate a frozen model is by using``freeze`` factory function.
    """

    def __init__(self, model, a0, *args):
        super().__init__(model, *(a0, *args))

    @property
    def a0(self):
        return self.args[0]

    @a0.setter
    def a0(self, value):
        self.args = (value,) + self.args[1:]



import numpy as np
import pytest

from relife.lifetime_model import Exponential
from relife.policy import OneCycleRunToFailurePolicy, RunToFailurePolicy


class TestOneCycleRunToFailure:
    def test_asymptotic_expected_equivalent_annual_cost(self, distribution, cf, discounting_rate):
        if isinstance(distribution, Exponential):
            pytest.skip("Exponential distribution won't work with this cf (not tested in v1.0.0 too)")
        policy = OneCycleRunToFailurePolicy(distribution, cf, discounting_rate=discounting_rate)
        qa = policy.asymptotic_expected_equivalent_annual_cost()  # () or (m,)
        assert qa.shape == cf.shape

    def test_expected_equivalent_annual_cost(self, distribution, cf, discounting_rate):
        if isinstance(distribution, Exponential):
            pytest.skip("Exponential distribution won't work with this cf (not tested in v1.0.0 too)")
        policy = RunToFailurePolicy(distribution, cf, discounting_rate=discounting_rate)
        qa = policy.asymptotic_expected_equivalent_annual_cost()
        timeline, q = policy.expected_equivalent_annual_cost(400, nb_steps=2000)
        assert timeline.shape == (2000,)
        assert q.shape == qa.shape + timeline.shape


class TestRunToFailure:
    def test_asymptotic_expected_equivalent_annual_cost(self, distribution, cf, discounting_rate):
        if isinstance(distribution, Exponential):
            pytest.skip("Exponential distribution won't work with this cf (not tested in v1.0.0 too)")
        policy = OneCycleRunToFailurePolicy(distribution, cf, discounting_rate=discounting_rate)
        qa = policy.asymptotic_expected_equivalent_annual_cost()  # () or (m,)
        assert qa.shape == cf.shape

    def test_expected_equivalent_annual_cost(self, distribution, cf, discounting_rate):
        if isinstance(distribution, Exponential):
            pytest.skip("Exponential distribution won't work with this cf (not tested in v1.0.0 too)")
        policy = RunToFailurePolicy(distribution, cf, discounting_rate=discounting_rate)
        qa = policy.asymptotic_expected_equivalent_annual_cost()
        timeline, q = policy.expected_equivalent_annual_cost(400, nb_steps=2000)
        assert timeline.shape == (2000,)
        assert q.shape == qa.shape + timeline.shape
        assert q[..., -1].flatten() == pytest.approx(qa.flatten(), rel=1e-1)


# FIXME : does not work because now max ndim in ls_integrate is 2d, here it's 3d -> broadcasting error
# possible solutions :
# 1. skip this test (obsolete) (-> recommanded)
# 2. set ndim (ReLife 1)
# 3. create a more complex mecanism to infer ndim inside ls_integrate (future improvements?)
# def test_expected_total_cost_vec(policy_vec):
#     batch_size = 3
#     timeline = np.arange(0, 100, 0.5)
#     z = policy_vec.expected_total_cost(timeline)
#     assert z.sum(axis=0) == pytest.approx(batch_size * z[0, ...], rel=1e-4)


# FIXME : does not work for end_time == 0.
#  In ReLife 1, when end_time == 0., atleast one sample was returned.
# possible solutions :
# 1. skip this test (obsolete) (-> recommanded), return one sample with end_time 0. has no sense
# 2.  change still_valid update order in sample_routine ?
# def test_sample(policy):
#     nb_assets = 5  # supposed to be set at initialization
#     nb_samples = 10
#     if isinstance(policy, RunToFailure):
#         data = policy.sample(nb_samples, 0.0)
#     else:
#         data = policy.sample(nb_samples)
#     assert len(data) == nb_samples * nb_assets



import numpy as np
from pytest import approx

from relife import ParametricModel


class ModelA(ParametricModel):
    def __init__(self, x, y):
        super().__init__(x=x, y=y)


class ModelB(ParametricModel):
    def __init__(self, model: ModelA, coef: tuple[float, ...]):
        super().__init__(**{f"coef_{i+1}": v for i, v in enumerate(coef)})
        self.baseline = model


def test_model_composition():
    model_a = ModelA(1, 2)
    assert model_a.params == approx(np.array([1, 2], dtype=np.float64))
    assert model_a.params_names == ("x", "y")

    model_b = ModelB(model_a, (3, 4, 5))
    assert model_b.params == approx(np.array([3, 4, 5, 1, 2], dtype=np.float64))
    assert model_b.params_names == ("coef_1", "coef_2", "coef_3", "x", "y")

    model_a.params = np.array([2, 3])
    assert model_a.params == approx(np.array([2, 3], dtype=np.float64))

    model_b.params = np.array([2, 3, 4, 5, 6])
    assert model_b.params == approx(np.array([2, 3, 4, 5, 6], dtype=np.float64))

    assert model_b.baseline.params == approx(np.array([5, 6], dtype=np.float64))

    model_b.baseline.params = np.array([1, 2])
    assert model_b.params == approx(np.array([2, 3, 4, 1, 2], dtype=np.float64))



from __future__ import annotations

from typing import TYPE_CHECKING, Optional

import numpy as np
from numpy.typing import NDArray

from relife.economic import RunToFailureReward
from relife.stochastic_process import RenewalRewardProcess

from ._base import BaseAgeReplacementPolicy, BaseOneCycleAgeReplacementPolicy

if TYPE_CHECKING:
    from relife.lifetime_model import (
        FrozenLeftTruncatedModel,
        LeftTruncatedModel,
    )

    from ..lifetime_model.distribution import LifetimeDistribution
    from ..lifetime_model.regression import FrozenLifetimeRegression


class OneCycleRunToFailurePolicy(BaseOneCycleAgeReplacementPolicy):
    # noinspection PyUnresolvedReferences
    r"""One cyle run-to-failure policy.

    Parameters
    ----------
    lifetime_model : any lifetime distribution or frozen lifetime model
        A lifetime model representing the durations between events.
    cf : float or 1darray
        Costs of failures
    discounting_rate : float, default is 0.
        The discounting rate value used in the exponential discounting function
    a0 : float or 1darray, optional
        Current ages of the assets, by default 0 for each asset. If it is given, left truncations of ``a0`` will
        be take into account for the first cycle.

    Attributes
    ----------
    cf
    """

    reward: RunToFailureReward

    def __init__(
        self,
        lifetime_model: LifetimeDistribution | FrozenLifetimeRegression | FrozenLeftTruncatedModel,
        cf: float | NDArray[np.float64],
        discounting_rate: float = 0.0,
        period_before_discounting: float = 1.0,
        a0: Optional[float | NDArray[np.float64]] = None,
    ) -> None:
        reward = RunToFailureReward(cf)
        if a0 is not None:
            lifetime_model = LeftTruncatedModel(lifetime_model).freeze(a0)
        super().__init__(lifetime_model, reward, discounting_rate, period_before_discounting)

    @property
    def cf(self):
        return self.reward.cf


class RunToFailurePolicy(BaseAgeReplacementPolicy):
    # noinspection PyUnresolvedReferences
    r"""Run-to-failure renewal policy.

    Renewal reward stochastic_process where assets are replaced on failure with costs
    :math:`c_f`.

    Parameters
    ----------
    lifetime_model : any lifetime distribution or frozen lifetime model
        A lifetime model representing the durations between events.
    cf : float or 1darray
        Costs of failures
    discounting_rate : float, default is 0.
        The discounting rate value used in the exponential discounting function
    first_lifetime_model : any lifetime distribution or frozen lifetime model, optional
        A lifetime model for the first renewal (delayed renewal process). It is None by default
    a0 : float or 1darray, optional
        Current ages of the assets, by default 0 for each asset. If it is given, left truncations of ``a0`` will
        be take into account for the first cycle.

    Attributes
    ----------
    cf

    References
    ----------
    .. [1] Van der Weide, J. A. M., & Van Noortwijk, J. M. (2008). Renewal
        theory with exponential and hyperbolic discounting. Probability in
        the Engineering and Informational Sciences, 22(1), 53-74.
    """

    lifetime_model: LifetimeDistribution | FrozenLifetimeRegression
    first_lifetime_model: Optional[LifetimeDistribution | FrozenLifetimeRegression | FrozenLeftTruncatedModel]
    reward: RunToFailureReward
    first_reward: Optional[RunToFailureReward]

    def __init__(
        self,
        lifetime_model: LifetimeDistribution | FrozenLifetimeRegression,
        cf: float | NDArray[np.float64],
        discounting_rate: float = 0.0,
        first_lifetime_model: Optional[
            LifetimeDistribution | FrozenLifetimeRegression | FrozenLeftTruncatedModel
        ] = None,
        a0: Optional[float | NDArray[np.float64]] = None,
    ) -> None:

        reward = RunToFailureReward(cf)
        first_reward: Optional[RunToFailureReward] = None
        if first_lifetime_model is not None:
            if a0 is not None:
                first_lifetime_model: FrozenLeftTruncatedModel = LeftTruncatedModel(first_lifetime_model).freeze(a0)
            first_reward = RunToFailureReward(cf)
        elif a0 is not None:
            first_lifetime_model = LeftTruncatedModel(lifetime_model).freeze(a0)

        stochastic_process = RenewalRewardProcess(
            lifetime_model,
            reward,
            discounting_rate,
            first_lifetime_model=first_lifetime_model,
            first_reward=first_reward,
        )
        super().__init__(stochastic_process)

    @property
    def cf(self):
        """
        Cost of failures

        Returns
        -------
        ndarray
        """
        return self.reward.cf

    @cf.setter
    def cf(self, value: float | NDArray[np.float64]) -> None:
        self.stochastic_process.reward.cf = value
        self.stochastic_process.first_reward.cf = value


from ._docstring import (
    ASYMPTOTIC_EEAC_DOCSTRING,
    ASYMPTOTIC_ETC_DOCSTRING,
    EEAC_DOCSTRING,
    ETC_DOCSTRING,
)

OneCycleRunToFailurePolicy.expected_total_cost.__doc__ = ETC_DOCSTRING
OneCycleRunToFailurePolicy.expected_equivalent_annual_cost.__doc__ = EEAC_DOCSTRING
OneCycleRunToFailurePolicy.asymptotic_expected_total_cost.__doc__ = ASYMPTOTIC_ETC_DOCSTRING
OneCycleRunToFailurePolicy.asymptotic_expected_equivalent_annual_cost.__doc__ = ASYMPTOTIC_EEAC_DOCSTRING

RunToFailurePolicy.expected_total_cost.__doc__ = ETC_DOCSTRING
RunToFailurePolicy.expected_equivalent_annual_cost.__doc__ = EEAC_DOCSTRING
RunToFailurePolicy.asymptotic_expected_total_cost.__doc__ = ASYMPTOTIC_ETC_DOCSTRING
RunToFailurePolicy.asymptotic_expected_equivalent_annual_cost.__doc__ = ASYMPTOTIC_EEAC_DOCSTRING



from __future__ import annotations

import copy
from dataclasses import InitVar, dataclass, field
from itertools import zip_longest
from typing import NamedTuple, Optional, TypeVarTuple

import numpy as np
from numpy.typing import NDArray


def time_reshape(time: NDArray[np.float64]) -> NDArray[np.float64]:
    # Check time shape
    if time.ndim > 2 or (time.ndim == 2 and time.shape[-1] not in (1, 2)):
        raise ValueError(f"Invalid time shape, got {time.shape} be time must be (m,), (m, 1) or (m,2)")
    if time.ndim < 2:
        time = time.reshape(-1, 1)  # time is (m, 1) or (m, 2)
    return time


def event_reshape(event: Optional[NDArray[np.bool_]] = None) -> Optional[NDArray[np.bool_]]:
    if event is not None:
        if event.ndim > 2 or (event.ndim == 2 and event.shape[-1] != 1):
            raise ValueError(f"Invalid event shape, got {event.shape} be event must be (m,) or (m, 1)")
        if event.ndim < 2:
            event = event.reshape(-1, 1)
        return event
    return None


def entry_reshape(entry: Optional[NDArray[np.float64]] = None) -> Optional[NDArray[np.float64]]:
    if entry is not None:
        if entry.ndim > 2 or (entry.ndim == 2 and entry.shape[-1] != 1):
            raise ValueError(f"Invalid entry shape, got {entry.shape} be entry must be (m,) or (m, 1)")
        if entry.ndim < 2:
            entry = entry.reshape(-1, 1)
        return entry
    return None


def departure_reshape(departure: Optional[NDArray[np.float64]] = None) -> Optional[NDArray[np.float64]]:
    if departure is not None:
        if departure.ndim > 2 or (departure.ndim == 2 and departure.shape[-1] != 1):
            raise ValueError(f"Invalid departure shape, got {departure.shape} be departure must be (m,) or (m, 1)")
        if departure.ndim < 2:
            departure = departure.reshape(-1, 1)
        return departure
    return None


def args_reshape(args: tuple[float | NDArray[np.float64], ...] = ()) -> tuple[NDArray[np.float64], ...]:
    args_list: list[NDArray[np.float64]] = [np.asarray(arg) for arg in args]
    for i, arg in enumerate(args_list):
        if arg.ndim > 2:
            raise ValueError(f"Invalid arg shape, got {arg.shape} shape at position {i}")
        if arg.ndim < 2:
            args_list[i] = arg.reshape(-1, 1)
    return tuple(args_list)


def get_complete(
    time: NDArray[np.float64], event: Optional[NDArray[np.bool_]] = None, args: tuple[NDArray[np.float64], ...] = ()
) -> Optional[IndexedLifetimeData]:
    if time.shape[-1] == 1:  # 1D time
        if event is None:
            event = np.ones((len(time), 1)).astype(np.bool_)
        index = np.where(event)[0]
        if index.size > 0:
            values = time[index]  # (m, 1)
            args = tuple((arg[index] for arg in args))
            return IndexedLifetimeData(values, index, args)
        return None
    else:  # 2D time
        index = np.where(time[:, 0] == time[:, 1])[0]
        if index.size > 0:
            values = time[index][:, [0]]  # (m, 1)
            args = tuple((arg[index] for arg in args))
            return IndexedLifetimeData(values, index, args)
        return None


def get_left_censoring(
    time: NDArray[np.float64], args: tuple[NDArray[np.float64], ...] = ()
) -> Optional[IndexedLifetimeData]:
    if time.shape[-1] == 1:  # 1D time
        return None
    else:  # 2D time
        index = np.where(time[:, 0] == 0)[0]
        if index.size > 0:
            values = time[index][:, [1]]  # (m, 1)
            args = tuple((arg[index] for arg in args))
            return IndexedLifetimeData(values, index, args)
        return None


def get_right_censoring(
    time: NDArray[np.float64], event: Optional[NDArray[np.bool_]] = None, args: tuple[NDArray[np.float64], ...] = ()
) -> Optional[IndexedLifetimeData]:
    if time.shape[-1] == 1:  # 1D time
        if event is None:
            event = np.ones((len(time), 1)).astype(np.bool_)
        index = np.where(~event)[0]
        if index.size > 0:
            values = time[index]  # (m, 1)
            args = tuple((arg[index] for arg in args))
            return IndexedLifetimeData(values, index, args)
        return None
    else:  # 2D time
        index = np.where(time[:, 1] == np.inf)[0]
        if index.size > 0:
            values = time[index][:, [0]]  # (m, 1)
            args = tuple((arg[index] for arg in args))
            return IndexedLifetimeData(values, index, args)
        return None


def get_interval_censoring(
    time: NDArray[np.float64], event: Optional[NDArray[np.bool_]] = None, args: tuple[NDArray[np.float64], ...] = ()
) -> Optional[IndexedLifetimeData]:
    if time.shape[-1] == 1:  # 1D time
        if event is None:
            event = np.ones((len(time), 1)).astype(np.bool_)
        rc_index = np.where(~event)[0]
        if rc_index.size > 0:
            rc_values = np.c_[time[rc_index], np.ones(len(rc_index)) * np.inf]  # add a column of inf
            args = tuple((arg[rc_index] for arg in args))
            return IndexedLifetimeData(rc_values, rc_index, args)
        return None
    else:  # 2D time
        index = np.where(
            np.not_equal(time[:, 0], time[:, 1]),
        )[0]
        if index.size > 0:
            values = time[index]  # (m, 2)
            if values.size != 0:
                if np.any(values[:, 0] >= values[:, 1]):
                    raise ValueError("Interval censorships lower bounds can't be higher or equal to its upper bounds")
            args = tuple((arg[index] for arg in args))
            return IndexedLifetimeData(values, index, args)
        return None


def get_left_truncation(
    time: NDArray[np.float64], entry: Optional[NDArray[np.float64]] = None, args: tuple[NDArray[np.float64], ...] = ()
) -> Optional[IndexedLifetimeData]:
    if entry is None:
        entry = np.zeros((len(time), 1))
    if time.shape[-1] == 1:  # 1D time
        index = np.where(entry > 0)[0]
        if index.size > 0:
            values = entry[index]  # (m, 1)
            args = tuple((arg[index] for arg in args))
            return IndexedLifetimeData(values, index, args)
        return None
    else:  # 2D time
        index = np.where(entry > 0)[0]
        if index.size > 0:
            values = entry[index]  # (m, 1)
            args = tuple((arg[index] for arg in args))
            return IndexedLifetimeData(values, index, args)
        return None


def get_right_truncation(
    time: NDArray[np.float64],
    departure: Optional[NDArray[np.float64]] = None,
    args: tuple[NDArray[np.float64], ...] = (),
) -> Optional[IndexedLifetimeData]:
    if departure is None:
        departure = np.ones((len(time), 1)) * np.inf
    if time.shape[-1] == 1:  # 1D time
        index = np.where(departure < np.inf)[0]
        if index.size > 0:
            values = departure[index]  # (m, 1)
            args = tuple((arg[index] for arg in args))
            return IndexedLifetimeData(values, index, args)
        return None
    else:  # 2D time
        index = np.where(departure < np.inf)[0]
        if index.size > 0:
            values = departure[index]  # (m, 1)
            args = tuple((arg[index] for arg in args))
            return IndexedLifetimeData(values, index, args)
        return None


Args = TypeVarTuple("Args")


class IndexedLifetimeData(NamedTuple):
    """
    Object that encapsulates lifetime data values and corresponding units index
    """

    lifetime_values: NDArray[np.float64]  # (m, 1) or (m, 2)
    lifetime_index: NDArray[np.int64]  # (m,)
    args: tuple[NDArray[np.float64], ...] = ()


@dataclass
class LifetimeData:
    nb_samples: int = field(init=False)

    complete: Optional[IndexedLifetimeData] = field(init=False, repr=False)  # values shape (m, 1)
    right_censoring: Optional[IndexedLifetimeData] = field(init=False, repr=False)  # values shape (m, 1)
    left_censoring: Optional[IndexedLifetimeData] = field(init=False, repr=False)  # values shape (m, 1)
    interval_censoring: Optional[IndexedLifetimeData] = field(init=False, repr=False)  # values shape (m, 2)
    left_truncation: Optional[IndexedLifetimeData] = field(init=False, repr=False)  # values shape (m, 1)
    right_truncation: Optional[IndexedLifetimeData] = field(init=False, repr=False)  # values shape (m, 1)
    complete_or_right_censored: Optional[IndexedLifetimeData] = field(init=False, repr=False)  # values shape (m, 1)

    time: InitVar[NDArray[np.float64]]
    event: InitVar[Optional[NDArray[np.bool_]]] = None
    entry: InitVar[Optional[NDArray[np.float64]]] = None
    departure: InitVar[Optional[NDArray[np.float64]]] = None
    args: InitVar[tuple[*Args]] = ()

    def __len__(self):
        return self.nb_samples

    def __post_init__(self, time, event, entry, departure, args):
        time = time_reshape(time)
        event = event_reshape(event)
        if time.shape[1] == 2:
            if event is not None:
                raise ValueError(
                    "When time is 2d, event is not necessary because time already encodes event information. Remove event"
                )
        entry = entry_reshape(entry)
        departure = departure_reshape(departure)
        args = args_reshape(args)

        # Check sizes
        sizes = [len(x) for x in (time, event, entry, departure, *args) if x is not None]
        if len(set(sizes)) != 1:
            raise ValueError(
                f"All lifetime data must have the same number of values. Fields length are different. Got {set(sizes)}"
            )

        # TODO : control here
        self.nb_samples = len(time)
        self.complete = get_complete(time, event, args)
        self.right_censoring = get_right_censoring(time, event, args)
        self.left_censoring = get_left_censoring(time, args)
        self.interval_censoring = get_interval_censoring(time, event, args)
        self.left_truncation = get_left_truncation(time, entry, args)
        self.right_truncation = get_right_truncation(time, departure, args)

        # sanity checks that observed lifetimes are inside truncation bounds
        for field_name in [
            "complete",
            "left_censoring",
            "left_censoring",
            "interval_censoring",
        ]:
            data = getattr(self, field_name)
            if data is not None and self.left_truncation is not None:
                inter_ids = np.intersect1d(data.lifetime_index, self.left_truncation.lifetime_index)
                intersection_values = np.concatenate(
                    (
                        data.lifetime_values[np.isin(data.lifetime_index, inter_ids)],
                        self.left_truncation.lifetime_values[np.isin(self.left_truncation.lifetime_index, inter_ids)],
                    ),
                    axis=1,
                )
                if len(intersection_values) != 0:
                    if np.any(
                        # take right bound when left bound is 0, otherwise take the min value of the bounds
                        # for none interval lifetimes, min equals the value
                        np.where(
                            intersection_values[:, [0]] == 0,
                            intersection_values[:, [-2]],
                            np.min(intersection_values[:, :-1], axis=1, keepdims=True),
                            # min of all cols but last
                        )
                        < intersection_values[:, [-1]]  # then check if any is under left truncation bound
                    ):
                        raise ValueError("Some lifetimes are under left truncation bounds")
            if data is not None and self.right_truncation is not None:
                inter_ids = np.intersect1d(data.lifetime_index, self.right_truncation.lifetime_index)
                intersection_values = np.concatenate(
                    (
                        data.lifetime_values[np.isin(data.lifetime_index, inter_ids)],
                        self.right_truncation.lifetime_values[np.isin(self.right_truncation.lifetime_index, inter_ids)],
                    ),
                    axis=1,
                )

                if len(intersection_values) != 0:
                    if np.any(
                        # take left bound when right bound is inf, otherwise take the max value of the bounds
                        # for none interval lifetimes, max equals the value
                        np.where(
                            intersection_values[:, [-2]] == np.inf,
                            intersection_values[:, [0]],
                            np.max(intersection_values[:, :-1], axis=1, keepdims=True),
                            # max of all cols but last
                        )
                        > intersection_values[:, [-1]]  # then check if any is above right truncation bound
                    ):
                        raise ValueError("Some lifetimes are above right truncation bounds")

        # compute complete_or_right_censored
        self.complete_or_right_censored = None
        if self.complete is not None and self.right_censoring is not None:
            values = np.concatenate([self.complete.lifetime_values, self.right_censoring.lifetime_values], axis=0)
            index = np.concatenate([self.complete.lifetime_index, self.right_censoring.lifetime_index], axis=0)
            args = tuple(
                (np.concatenate(x, axis=0) for x in zip_longest(self.complete.args, self.right_censoring.args))
            )
            # FIXME: orders of the values seems to affects estimations of the parameters in Regression
            sort_index = np.argsort(index)

            self.complete_or_right_censored = IndexedLifetimeData(
                values[sort_index],
                index[sort_index],
                tuple((arg[sort_index] for arg in args)),
            )
        elif self.complete is not None:
            self.complete_or_right_censored = copy.deepcopy(self.complete)
        elif self.right_censoring is not None:
            self.complete_or_right_censored = copy.deepcopy(self.right_censoring)



from __future__ import annotations

from typing import TYPE_CHECKING, Generic, Optional, TypeVarTuple

import matplotlib.pyplot as plt
import numpy as np
import scipy.stats as stats
from matplotlib.axes import Axes
from numpy.ma.core import zeros_like
from numpy.typing import NDArray

if TYPE_CHECKING:
    from relife.lifetime_model import (
        NonParametricLifetimeModel,
        ParametricLifetimeModel,
    )

ALPHA_CI: float = 0.05


def plot_prob_function(
    x: NDArray[np.float64],
    y: NDArray[np.float64],
    se: Optional[NDArray[np.float64]] = None,
    ci_bounds: Optional[tuple[float, float]] = None,
    label: Optional[str] = None,
    drawstyle: str = "default",
    **kwargs,
) -> Axes:
    ax = kwargs.pop("ax", plt.gca())
    ax.plot(x, y, drawstyle=drawstyle, label=label, **kwargs)
    if se is not None and ci_bounds is not None:
        z = stats.norm.ppf(1 - ALPHA_CI / 2)
        yl = np.clip(y - z * se, ci_bounds[0], ci_bounds[1])
        yu = np.clip(y + z * se, ci_bounds[0], ci_bounds[1])
        step = drawstyle.split("-")[1] if "steps-" in drawstyle else None
        ax.fill_between(x, yl, yu, facecolors=[ax.lines[-1].get_color()], step=step, alpha=0.25, label="IC-95%")
    if label is not None:
        ax.legend()
    return ax


Args = TypeVarTuple("Args")


class PlotParametricLifetimeModel(Generic[*Args]):
    def __init__(self, model: ParametricLifetimeModel[*Args]):
        self.model = model

    def _plot(self, fname: str, *args: *Args, ci_bounds: Optional[tuple[float, float]] = None, **kwargs) -> Axes:
        end_time = kwargs.pop("end_time", None)
        if end_time is None:
            end_time = np.squeeze(self.model.isf(1e-3, *args))  # () or (m,)
        timeline = np.linspace(0, end_time, 200, dtype=np.float64)  # (200,) or (200, m)
        timeline = np.transpose(timeline)  # (200,) (m, 200)
        f = getattr(self.model, fname)
        jac_f = getattr(self.model, "jac_" + fname, None)
        y = f(timeline, *args)
        se = None
        if getattr(self.model, "fitting_results", None) is not None and jac_f is not None:
            se = zeros_like(timeline)
            se[..., 1:] = self.model.fitting_results.se_estimation_function(
                jac_f(timeline[..., 1:], *args, asarray=True)
            )
        label = kwargs.pop("label", f"{self.model.__class__.__name__}" + f".{fname}")
        ax = plot_prob_function(timeline, y, se=se, ci_bounds=ci_bounds, label=label, **kwargs)
        return ax

    def sf(self, *args: *Args, **kwargs) -> Axes:
        end_time = kwargs.get("end_time", None)
        ax = self._plot("sf", *args, ci_bounds=(0.0, 1.0), **kwargs)
        ax.set_ylabel("$S(t) = P(T > t)$")
        ax.set_xlabel("t")
        ax.set_title("Survival function")
        ax.set_ylim(bottom=0)
        ax.set_xlim(left=0, right=end_time)
        return ax

    def cdf(self, *args: *Args, **kwargs) -> Axes:
        end_time = kwargs.get("end_time", None)
        ax = self._plot("cdf", *args, ci_bounds=(0.0, 1.0), **kwargs)
        ax.set_ylabel("$F(t) = P(T \leq t)$")
        ax.set_xlabel("t")
        ax.set_title("Cumulative distribution function")
        ax.set_ylim(bottom=0)
        ax.set_xlim(left=0, right=end_time)
        return ax

    def chf(self, *args: *Args, **kwargs) -> Axes:
        end_time = kwargs.get("end_time", None)
        ax = self._plot("chf", *args, ci_bounds=(0.0, np.inf), **kwargs)
        ax.set_ylabel("$H(t)$")
        ax.set_xlabel("t")
        ax.set_title("Cumulative hazard function")
        ax.set_ylim(bottom=0)
        ax.set_xlim(left=0, right=end_time)
        return ax

    def hf(self, *args: *Args, **kwargs) -> Axes:
        end_time = kwargs.get("end_time", None)
        ax = self._plot("hf", *args, ci_bounds=(0.0, np.inf), **kwargs)
        ax.set_ylabel("$h(t)$")
        ax.set_xlabel("t")
        ax.set_title("Hazard function")
        ax.set_ylim(bottom=0)
        ax.set_xlim(left=0, right=end_time)
        return ax

    def pdf(self, *args: *Args, **kwargs) -> Axes:
        end_time = kwargs.get("end_time", None)
        ax = self._plot("pdf", *args, ci_bounds=(0.0, np.inf), **kwargs)
        ax.set_ylabel("$f(t)$")
        ax.set_xlabel("t")
        ax.set_title("Probability density function")
        ax.set_ylim(bottom=0)
        ax.set_xlim(left=0, right=end_time)
        return ax


class PlotNonParametricLifetimeModel:
    def __init__(self, model: NonParametricLifetimeModel):
        self.model = model

    def plot(self, fname: str, plot_se: bool = True, ci_bounds=(0.0, 1.0), drawstyle="steps-post", **kwargs) -> Axes:
        label = kwargs.pop("label", f"{self.model.__class__.__name__}" + f".{fname}")
        res = getattr(self.model, fname)(se=plot_se)
        se = None if not plot_se else res[-1]
        timeline, y = res[:2]
        ax = plot_prob_function(timeline, y, se=se, ci_bounds=ci_bounds, label=label, drawstyle=drawstyle, **kwargs)
        return ax


class PlotECDF(PlotNonParametricLifetimeModel):
    def sf(self, plot_se: bool = True, **kwargs) -> Axes:
        ax = self.plot("sf", plot_se=plot_se, **kwargs)
        ax.set_ylabel("Estimated survival function")
        ax.set_xlabel("Time")
        ax.set_ylim(bottom=0)
        ax.set_xlim(left=0)
        return ax

    def cdf(self, plot_se: bool = True, **kwargs) -> Axes:
        ax = self.plot("cdf", plot_se=plot_se, **kwargs)
        ax.set_ylabel("Estimated cumulative distribution function")
        ax.set_xlabel("Time")
        ax.set_ylim(bottom=0)
        ax.set_xlim(left=0)
        return ax


class PlotKaplanMeier(PlotNonParametricLifetimeModel):
    def sf(self, plot_se: bool = True, **kwargs) -> Axes:
        ax = self.plot("sf", plot_se=plot_se, **kwargs)
        ax.set_ylabel("Estimated survival function")
        ax.set_xlabel("Time")
        ax.set_ylim(bottom=0)
        ax.set_xlim(left=0)
        return ax


class PlotNelsonAalen(PlotNonParametricLifetimeModel):
    def chf(self, plot_se: bool = True, **kwargs) -> Axes:
        ax = self.plot("chf", plot_se=plot_se, ci_bounds=(0.0, np.inf), **kwargs)
        ax.set_ylabel("Estimated cumulative hazard function")
        ax.set_xlabel("Time")
        ax.set_ylim(bottom=0)
        ax.set_xlim(left=0)
        return ax


class PlotTurnbull(PlotNonParametricLifetimeModel):
    def sf(self, **kwargs) -> Axes:
        ax = self.plot("sf", plot_se=False, **kwargs)
        ax.set_ylabel("Estimated survival function")
        ax.set_xlabel("Time")
        ax.set_xlim(left=0)
        return ax


# def count_data_plot(
#     fname: str,
#     obj: CountData,
#     **kwargs,
# ):
#     label = kwargs.pop("label", fname)
#     if not hasattr(obj, fname):
#         raise ValueError(f"No plot for {fname}")
#     timeline, values = getattr(obj, fname)()
#     return plot(timeline, values, drawstyle="steps-post", label=label, **kwargs)
#
#
# def nhpp_count_data_plot(
#     fname: str,
#     obj: NHPPCountData,
#     **kwargs,
# ):
#     label = kwargs.pop("label", fname)
#     if not hasattr(obj, fname):
#         raise ValueError(f"No plot for {fname}")
#     timeline, values = getattr(obj, fname)()
#     if fname in ("total_rewards", "mean_total_rewards"):
#         ax = kwargs.pop("ax", plt.gca())
#         alpha = kwargs.pop("alpha", 0.2)
#         ax.plot(timeline, values, drawstyle="steps-post", label=label, **kwargs)
#         ax.fill_between(timeline, values, where=values >= 0, step="post", alpha=alpha, **kwargs)
#         if label is not None:
#             ax.legend()
#         return ax
#     return plot(timeline, values, drawstyle="steps-post", label=label, **kwargs)
#
#
# def renewal_data_plot(
#     fname: str,
#     obj: CountData,
#     **kwargs,
# ):
#     label = kwargs.pop("label", fname)
#     if not hasattr(obj, fname):
#         raise ValueError(f"No plot for {fname}")
#     timeline, values = getattr(obj, fname)()
#     if fname in ("total_rewards", "mean_total_rewards"):
#         ax = kwargs.pop("ax", plt.gca())
#         alpha = kwargs.pop("alpha", 0.2)
#         ax.plot(timeline, values, drawstyle="steps-post", label=label, **kwargs)
#         ax.fill_between(timeline, values, where=values >= 0, step="post", alpha=alpha, **kwargs)
#         if label is not None:
#             ax.legend()
#         return ax
#     else:
#         return count_data_plot(fname, obj, label=label, **kwargs)


# def nhpp_plot(
#     fname: str,
#     obj: NonHomogeneousPoissonProcess,
#     timeline: NDArray[np.float64] = None,
#     **kwargs,
# ):
#
#     label = kwargs.pop("label", f"{obj.__class__.__name__}" + f".{fname}")
#     if not hasattr(obj, fname):
#         raise ValueError(f"No plot for {fname}")
#     f = getattr(obj, fname)
#     y = f(timeline)
#     return plot(timeline, y, label=label, **kwargs)



from typing import Generic

from relife import FrozenParametricModel, ParametricModel
from relife._typing import _Xs

class StochasticProcess(ParametricModel, Generic[*_Xs]): ...

class FrozenStochasticProcess(FrozenParametricModel[*_Xs]):
    model: StochasticProcess[*_Xs]
    args: tuple[*_Xs]
    def __init__(self, model: StochasticProcess[*_Xs], *args: *_Xs) -> None: ...



from abc import ABC
from typing import (
    Any,
    Callable,
    Literal,
    Optional,
    Self,
    overload,
)

import numpy as np
from numpy.typing import NDArray
from scipy.optimize import Bounds
from typing_extensions import override

from relife import ParametricModel as ParametricModel
from relife._typing import _B, _X, _Y, _Xs
from relife.lifetime_model import (
    FittableParametricLifetimeModel,
    FrozenParametricLifetimeModel,
)
from relife.likelihood import FittingResults as FittingResults

def broadcast_time_covar(time: _X, covar: _X) -> tuple[_X, _X]: ...
def broadcast_time_covar_shapes(
    time_shape: tuple[()] | tuple[int] | tuple[int, int], covar_shape: tuple[()] | tuple[int] | tuple[int, int]
) -> tuple[()] | tuple[int] | tuple[int, int]: ...

class LifetimeRegression(FittableParametricLifetimeModel[*tuple[_X, *_Xs]], ABC):
    fitting_results = Optional[FittingResults]
    covar_effect: CovarEffect
    baseline: FittableParametricLifetimeModel[*_Xs]

    def __init__(
        self,
        baseline: FittableParametricLifetimeModel[*_Xs],
        coefficients: tuple[Optional[float], ...] = (None,),
    ) -> None: ...
    def _get_initial_params(
        self,
        time: NDArray[np.float64],
        covar: NDArray[np.float64],
        *args: *_Xs,
        event: Optional[NDArray[np.bool_]] = None,
        entry: Optional[NDArray[np.float64]] = None,
        departure: Optional[NDArray[np.float64]] = None,
    ) -> NDArray[np.float64]: ...
    def _get_params_bounds(self) -> Bounds: ...
    @property
    def coefficients(self) -> NDArray[np.float64]: ...
    @property
    def nb_coef(self) -> int: ...
    @override
    def sf(self, time: _X, covar: _X, *args: *_Xs) -> _Y: ...
    @override
    def pdf(self, time: _X, covar: _X, *args: *_Xs) -> _Y: ...
    @override
    def isf(
        self,
        probability: _X,
        covar: _X,
        *args: *_Xs,
    ) -> _Y: ...
    @override
    def cdf(self, time: _X, covar: _X, *args: *_Xs) -> _Y: ...
    @override
    def ppf(
        self,
        probability: _X,
        covar: _X,
        *args: *_Xs,
    ) -> _Y: ...
    @override
    def mrl(self, time: _X, covar: _X, *args: *_Xs) -> _Y: ...
    @override
    def mean(self, covar: _X, *args: *_Xs) -> _Y: ...
    @override
    def var(self, covar: _X, *args: *_Xs) -> _Y: ...
    @override
    def median(self, covar: _X, *args: *_Xs) -> _Y: ...
    @override
    def ls_integrate(
        self,
        func: Callable[[_X], _Y],
        a: _X,
        b: _X,
        covar: _X,
        *args: *_Xs,
        deg: int = 10,
    ) -> _Y: ...
    @overload
    def jac_sf(
        self,
        time: _X,
        covar: _X,
        *args: *_Xs,
        asarray: Literal[False] = False,
    ) -> tuple[_Y, ...]: ...
    @overload
    def jac_sf(
        self,
        time: _X,
        covar: _X,
        *args: *_Xs,
        asarray: Literal[True] = True,
    ) -> _Y: ...
    def jac_sf(
        self,
        time: _X,
        covar: _X,
        *args: *_Xs,
        asarray: bool = True,
    ) -> tuple[_Y, ...] | _Y: ...
    @overload
    def jac_cdf(
        self,
        time: _X,
        covar: _X,
        *args: *_Xs,
        asarray: Literal[False] = False,
    ) -> tuple[_Y, ...]: ...
    @overload
    def jac_cdf(
        self,
        time: _X,
        covar: _X,
        *args: *_Xs,
        asarray: Literal[True] = True,
    ) -> _Y: ...
    def jac_cdf(
        self,
        time: _X,
        covar: _X,
        *args: *_Xs,
        asarray: bool = True,
    ) -> tuple[_Y, ...] | _Y: ...
    @overload
    def jac_pdf(
        self,
        time: _X,
        covar: _X,
        *args: *_Xs,
        asarray: Literal[False] = False,
    ) -> tuple[_Y, ...]: ...
    @overload
    def jac_pdf(
        self,
        time: _X,
        covar: _X,
        *args: *_Xs,
        asarray: Literal[True] = True,
    ) -> _Y: ...
    def jac_pdf(
        self,
        time: _X,
        covar: _X,
        *args: *_Xs,
        asarray: bool = True,
    ) -> tuple[_Y, ...] | _Y: ...
    @overload
    def rvs(
        self,
        size: int,
        covar: _X,
        *args: *_Xs,
        nb_assets: Optional[int] = None,
        return_event: Literal[False] = False,
        return_entry: Literal[False] = False,
        seed: Optional[int] = None,
    ) -> _Y: ...
    @overload
    def rvs(
        self,
        size: int,
        covar: _X,
        *args: *_Xs,
        nb_assets: Optional[int] = None,
        return_event: Literal[True] = True,
        return_entry: Literal[False] = False,
        seed: Optional[int] = None,
    ) -> tuple[_Y, _B]: ...
    @overload
    def rvs(
        self,
        size: int,
        covar: _X,
        *args: *_Xs,
        nb_assets: Optional[int] = None,
        return_event: Literal[False] = False,
        return_entry: Literal[True] = True,
        seed: Optional[int] = None,
    ) -> tuple[_Y, _Y]: ...
    @overload
    def rvs(
        self,
        size: int,
        covar: _X,
        *args: *_Xs,
        nb_assets: Optional[int] = None,
        return_event: Literal[True] = True,
        return_entry: Literal[True] = True,
        seed: Optional[int] = None,
    ) -> tuple[_Y, _B, _Y]: ...
    def rvs(
        self,
        size: int,
        covar: _X,
        *args: *_Xs,
        nb_assets: Optional[int] = None,
        return_event: bool = False,
        return_entry: bool = False,
        seed: Optional[int] = None,
    ) -> _Y | tuple[_Y, _Y] | tuple[_Y, _B] | tuple[_Y, _B, _Y]: ...
    def fit(
        self,
        time: NDArray[np.float64],
        covar: NDArray[np.float64],
        event: NDArray[np.bool_] | None = None,
        entry: NDArray[np.float64] | None = None,
        departure: NDArray[np.float64] | None = None,
        **options: Any,
    ) -> Self: ...
    def freeze(self, covar: _X, *args: *_Xs) -> FrozenLifetimeRegression[*_Xs]: ...

class ProportionalHazard(LifetimeRegression[*_Xs]):
    def hf(self, time: _X, covar: _X, *args: *_Xs) -> _Y: ...
    def chf(self, time: _X, covar: _X, *args: *_Xs) -> _Y: ...
    @override
    def ichf(
        self,
        cumulative_hazard_rate: _X,
        covar: _X,
        *args: *_Xs,
    ) -> _Y: ...
    def dhf(self, time: _X, covar: _X, *args: *_Xs) -> _Y: ...
    @overload
    def jac_hf(
        self,
        time: _X,
        covar: _X,
        *args: *_Xs,
        asarray: Literal[False] = False,
    ) -> tuple[_Y, ...]: ...
    @overload
    def jac_hf(
        self,
        time: _X,
        covar: _X,
        *args: *_Xs,
        asarray: Literal[True] = True,
    ) -> _Y: ...
    def jac_hf(
        self,
        time: _X,
        covar: _X,
        *args: *_Xs,
        asarray: bool = True,
    ) -> tuple[_Y, ...] | _Y: ...
    @overload
    def jac_chf(
        self,
        time: _X,
        covar: _X,
        *args: *_Xs,
        asarray: Literal[False] = False,
    ) -> tuple[_Y, ...]: ...
    @overload
    def jac_chf(
        self,
        time: _X,
        covar: _X,
        *args: *_Xs,
        asarray: Literal[True] = True,
    ) -> _Y: ...
    def jac_chf(
        self,
        time: _X,
        covar: _X,
        *args: *_Xs,
        asarray: bool = True,
    ) -> tuple[_Y, ...] | _Y: ...
    @override
    def moment(self, n: int, covar: _X, *args: *_Xs) -> _Y: ...

class AcceleratedFailureTime(LifetimeRegression[*_Xs]):
    def hf(self, time: _X, covar: _X, *args: *_Xs) -> _Y: ...
    def chf(self, time: _X, covar: _X, *args: *_Xs) -> _Y: ...
    @override
    def ichf(
        self,
        cumulative_hazard_rate: _X,
        covar: _X,
        *args: *_Xs,
    ) -> _Y: ...
    def dhf(self, time: _X, covar: _X, *args: *_Xs) -> _Y: ...
    @overload
    def jac_hf(
        self,
        time: _X,
        covar: _X,
        *args: *_Xs,
        asarray: Literal[False] = False,
    ) -> tuple[_Y, ...]: ...
    @overload
    def jac_hf(
        self,
        time: _X,
        covar: _X,
        *args: *_Xs,
        asarray: Literal[True] = True,
    ) -> _Y: ...
    def jac_hf(
        self,
        time: _X,
        covar: _X,
        *args: *_Xs,
        asarray: bool = True,
    ) -> tuple[_Y, ...] | _Y: ...
    @overload
    def jac_chf(
        self,
        time: _X,
        covar: _X,
        *args: *_Xs,
        asarray: Literal[False] = False,
    ) -> tuple[_Y, ...]: ...
    @overload
    def jac_chf(
        self,
        time: _X,
        covar: _X,
        *args: *_Xs,
        asarray: Literal[True] = True,
    ) -> _Y: ...
    def jac_chf(
        self,
        time: _X,
        covar: _X,
        *args: *_Xs,
        asarray: bool = True,
    ) -> tuple[_Y, ...] | _Y: ...
    @override
    def moment(self, n: int, covar: _X, *args: *_Xs) -> _Y: ...

class CovarEffect(ParametricModel):
    def __init__(self, coefficients: tuple[Optional[None], ...] = (None,)) -> None: ...
    @property
    def nb_coef(self) -> int: ...
    def g(self, covar: _X) -> _Y: ...
    @overload
    def jac_g(self, covar: _X, *, asarray: Literal[False] = False) -> tuple[_Y, ...]: ...
    @overload
    def jac_g(self, covar: _X, *, asarray: Literal[True] = True) -> _Y: ...
    def jac_g(self, covar: _X, *, asarray: Literal[True] = True) -> tuple[_Y, ...] | _Y: ...

class FrozenLifetimeRegression(FrozenParametricLifetimeModel[*tuple[_X, *_Xs]]):
    unfrozen_model: LifetimeRegression[*_Xs]
    args: tuple[_X, *_Xs]

    def __init__(self, model: LifetimeRegression[*_Xs], covar: _X, *args: *_Xs) -> None: ...
    @override
    def unfreeze(self) -> LifetimeRegression[*_Xs]: ...
    @property
    def nb_coef(self) -> int: ...
    @property
    def covar(self) -> _X: ...
    # noinspection PyUnresolvedReferences
    @covar.setter
    def covar(self, value: _X) -> None: ...
    def dhf(self, time: _X, *args: *_Xs) -> _Y: ...
    @overload
    def jac_hf(self, time: _X, asarray: Literal[False] = False) -> tuple[_Y, ...]: ...
    @overload
    def jac_hf(self, time: _X, asarray: Literal[True] = True) -> _Y: ...
    def jac_hf(self, time: _X, asarray: bool = True) -> tuple[_Y, ...] | _Y: ...
    @overload
    def jac_chf(self, time: _X, asarray: Literal[False] = False) -> tuple[_Y, ...]: ...
    @overload
    def jac_chf(self, time: _X, asarray: Literal[True] = True) -> _Y: ...
    def jac_chf(self, time: _X, asarray: bool = True) -> tuple[_Y, ...] | _Y: ...
    @overload
    def jac_sf(self, time: _X, asarray: Literal[False] = False) -> tuple[_Y, ...]: ...
    @overload
    def jac_sf(self, time: _X, asarray: Literal[True] = True) -> _Y: ...
    def jac_sf(self, time: _X, asarray: bool = True) -> tuple[_Y, ...] | _Y: ...
    @overload
    def jac_pdf(self, time: _X, asarray: Literal[False] = False) -> tuple[_Y, ...]: ...
    @overload
    def jac_pdf(self, time: _X, asarray: Literal[True] = True) -> _Y: ...
    def jac_pdf(self, time: _X, asarray: bool = True) -> tuple[_Y, ...] | _Y: ...



import numpy as np

from relife.data import LifetimeData, NHPPData
from relife.likelihood import LikelihoodFromLifetimes

from .base import FrozenStochasticProcess, StochasticProcess


class NonHomogeneousPoissonProcess(StochasticProcess):
    """
    Non-homogeneous Poisson process.
    """

    def __init__(self, lifetime_model):
        super().__init__()
        self.lifetime_model = lifetime_model
        self.fitting_results = None

    def intensity(self, time, *args):
        """
        The intensity function of the process.

        Parameters
        ----------
        time : float or np.ndarray
            Elapsed time value(s) at which to compute the function.
            If ndarray, allowed shapes are ``()``, ``(n,)`` or ``(m, n)``.
        *args : float or np.ndarray
            Additional arguments needed by the model.

        Returns
        -------
        np.float64 or np.ndarray
            Function values at each given time(s).
        """
        return self.lifetime_model.hf(time, *args)

    def cumulative_intensity(self, time, *args):
        """
        The cumulative intensity function of the process.

        Parameters
        ----------
        time : float or np.ndarray
            Elapsed time value(s) at which to compute the function.
            If ndarray, allowed shapes are ``()``, ``(n,)`` or ``(m, n)``.
        *args : float or np.ndarray
            Additional arguments needed by the model.

        Returns
        -------
        np.float64 or np.ndarray
            Function values at each given time(s).
        """
        return self.lifetime_model.chf(time, *args)

    def freeze(self, *args):
        """
        Freeze any arguments required by the process into the object data.

        Parameters
        ----------
        *args : float or np.ndarray
            Additional arguments needed by the model.

        Returns
        -------
        FrozenNonHomogeneousPoissonProcess
        """
        return FrozenNonHomogeneousPoissonProcess(self, *args)

    def sample(self, size, tf, *args, t0=0.0, seed=None):
        """Renewal data sampling.

        This function will sample data and encapsulate them in an object.

        Parameters
        ----------
        size : int
            The size of the desired sample
        *args : float or np.ndarray
            Additional arguments needed by the model.
        tf : float
            Time at the end of the observation.
        t0 : float, default 0
            Time at the beginning of the observation.
        seed : int, optional
            Random seed, by default None.

        """

        from ._sample import (
            NonHomogeneousPoissonProcessIterable,
            NonHomogeneousPoissonProcessSample,
        )

        frozen_nhpp = self.freeze(*args)
        iterable = NonHomogeneousPoissonProcessIterable(frozen_nhpp, size, tf, t0=t0, seed=seed)
        struct_array = np.concatenate(tuple(iterable))
        struct_array = np.sort(struct_array, order=("sample_id", "asset_id", "timeline"))
        return NonHomogeneousPoissonProcessSample(t0, tf, struct_array)

    def generate_failure_data(self, size, tf, *args, t0=0.0, seed=None):
        """Generate failure data

        This function will generate failure data that can be used to fit a non-homogeneous Poisson process.

        Parameters
        ----------
        size : int
            The size of the desired sample
        tf : float
            Time at the end of the observation.
        t0 : float, default 0
            Time at the beginning of the observation.
        seed : int, optional
            Random seed, by default None.

        Returns
        -------
        A dict of ages_at_events, events_assets_ids, first_ages, last_ages, model_args and assets_ids
        """
        from ._sample import NonHomogeneousPoissonProcessIterable

        frozen_nhpp = self.freeze(*args)

        iterable = NonHomogeneousPoissonProcessIterable(frozen_nhpp, size, tf, t0=t0, seed=seed)
        struct_array = np.concatenate(tuple(iterable))
        struct_array = np.sort(struct_array, order=("sample_id", "asset_id", "timeline"))

        first_ages_index = np.nonzero(struct_array["entry"] == t0)
        last_ages_index = np.nonzero(struct_array["age"] == tf)

        event_index = np.nonzero(struct_array["event"])

        first_ages = struct_array[first_ages_index]["entry"].copy()
        last_ages = struct_array[last_ages_index]["age"].copy()

        assets_ids = np.char.add(
            np.char.add(
                np.full_like(struct_array[last_ages_index]["sample_id"], "S", dtype=np.str_),
                struct_array[last_ages_index]["sample_id"].astype(np.str_),
            ),
            np.char.add(
                np.full_like(struct_array[last_ages_index]["asset_id"], "A", dtype=np.str_),
                struct_array[last_ages_index]["asset_id"].astype(np.str_),
            ),
        )

        events_assets_ids = np.char.add(
            np.char.add(
                np.full_like(struct_array[event_index]["sample_id"], "S", dtype=np.str_),
                struct_array[event_index]["sample_id"].astype(np.str_),
            ),
            np.char.add(
                np.full_like(struct_array[event_index]["asset_id"], "A", dtype=np.str_),
                struct_array[event_index]["asset_id"].astype(np.str_),
            ),
        )
        ages_at_events = struct_array[event_index]["age"].copy()

        return {
            "ages_at_events": ages_at_events,
            "events_assets_ids": events_assets_ids,
            "first_ages": first_ages,
            "last_ages": last_ages,
            "assets_ids": assets_ids,
        }

    def fit(
        self,
        ages_at_events,
        events_assets_ids,
        assets_ids=None,
        first_ages=None,
        last_ages=None,
        model_args=None,
        **options,
    ):
        """
        Estimation of the process parameters from recurrent failure data.

        Parameters
        ----------
        ages_at_events : 1D array of float
            Array of float containing the ages of each asset when the events occured
        events_assets_ids : sequence of hashable
            Sequence object containing the ids of each assets corresponding to each ages.
        first_ages : 1D array of float, optional
            Array of float containing the ages of each asset before observing events. If set, ``assets_ids`` is needed and its length
            must equal the size of ``first_ages``.
        last_ages : 1D array of float, optional
            Array of float containing the ages of each asset at the end of the observation period. If set, ``assets_ids`` is needed and its length
            must equal the size of ``last_ages``.
        model_args : tuple of np.ndarray, optional
            Additional arguments needed by the model. If set, ``assets_ids`` is needed.
            For 1D array, the size must equal the length of ``assets_ids``. For 2D array (e.g. covar of regression),
            the length of first axis must equal the length of ``assets_ids``.
        assets_ids : sequence of hashable, optional
            Only needed if either ``first_ages``, ``last_ages`` or ``model_args`` is filled. It must be a sequence object
            containing the unique ids corresponding to each values contained in ``first_ages``, ``last_ages`` and/or ``model_args``

        Returns
        -------
        Self
            The current object with the estimated parameters setted inplace.

        Examples
        --------

        Ages of assets AB2 and CX13 at each event.

        >>> from relife.lifetime_model import Weibull
        >>> from relife.stochastic_process import NonHomogeneousPoissonProcess
        >>> nhpp = NonHomogeneousPoissonProcess(Weibull())
        >>> nhpp.fit(
            np.array([11., 13., 21., 25., 27.]),
            ("AB2", "CX13", "AB2", "AB2", "CX13"),
        )

        With additional information and model args (regression of 2 coefficients)

        >>> from relife.lifetime_model import ProportionalHazard
        >>> nhpp = NonHomogeneousPoissonProcess(ProportionalHazard())
        >>> nhpp.fit(
            np.array([11., 13., 21., 25., 27.]),
            ("AB2", "CX13", "AB2", "AB2", "CX13"),
            first_ages = np.array([10., 12.]),
            last_ages = np.array([35., 60.]),
            model_args = (np.array([[1.2, 5.5], [37.2, 22.2]]),) # 2d array of 2 raws (2 assets) and 2 columns (2 coefficients)
        )
        """
        nhpp_data = NHPPData(
            ages_at_events,
            events_assets_ids,
            first_ages=first_ages,
            last_ages=last_ages,
            model_args=model_args,
            assets_ids=assets_ids,
        )
        time, event, entry, args = nhpp_data.to_lifetime_data()
        # noinspection PyProtectedMember
        self.lifetime_model._get_initial_params(time, *args, event=event, entry=entry)
        lifetime_data = LifetimeData(time, event=event, entry=entry, args=args)
        likelihood = LikelihoodFromLifetimes(self.lifetime_model, lifetime_data)
        fitting_results = likelihood.maximum_likelihood_estimation(**options)
        self.params = fitting_results.optimal_params
        self.fitting_results = fitting_results
        return self


class FrozenNonHomogeneousPoissonProcess(FrozenStochasticProcess):

    def intensity(self, time):
        return self.unfrozen_model.intensity(time, *self.args)

    def cumulative_intensity(self, time):
        return self.unfrozen_model.cumulative_intensity(time, *self.args)

    def sample(self, size, tf, t0=0.0, nb_assets=None, seed=None):
        return self.unfrozen_model.sample(size, tf, *self.args, t0=t0, nb_assets=nb_assets, seed=seed)

    def generate_failure_data(self, size, tf, t0=0.0, nb_assets=None, seed=None):
        return self.unfrozen_model.generate_failure_data(size, tf, t0=t0, nb_assets=nb_assets, seed=seed, *self.args)



import numpy as np
from pytest import approx

from relife.quadrature import laguerre_quadrature, legendre_quadrature


def test_laguerre_quadrature(integration_bound_a):
    # f : R -> R or R^n -> R^n or R^(m, n) -> R^(m, n), etc.
    f = lambda x: x
    # integral_a^inf x*exp(-x)dx = (a + 1)*exp(-a)
    expected_intg = lambda a: (a + 1) * np.exp(-a)

    integration = laguerre_quadrature(f, integration_bound_a)
    assert integration.shape == integration_bound_a.shape
    assert integration == approx(expected_intg(integration_bound_a))

    # g : R -> R^d or R^n -> R^(d, n) or R^(m, n) -> R^(d, m, n), etc.
    d = 6
    stack_f = lambda x: np.stack([x] * d, axis=0)
    integration = laguerre_quadrature(stack_f, integration_bound_a)
    assert integration.shape == (d,) + integration_bound_a.shape
    assert integration == approx(stack_f(expected_intg(integration_bound_a)))


def test_legendre_quadrature(integration_bound_a, integration_bound_b):
    m = 2
    n = 3

    # f : R -> R or R^n -> R^n or R^(m, n) -> R^(m, n), etc.
    f = lambda x: x
    # integral_a^b xdx = (1/2)*(b^2 - a^2)
    expected_intg = lambda a, b: 0.5 * (b**2 - a**2)

    integration = legendre_quadrature(f, integration_bound_a, integration_bound_b)
    assert integration.shape == np.broadcast_shapes(integration_bound_a.shape, integration_bound_b.shape)
    assert integration == approx(expected_intg(integration_bound_a, integration_bound_b))

    # f : R -> R^d or R^n -> R^(d, n) or R^(m, n) -> R^(d, m, n), etc.
    d = 6
    stack_f = lambda x: np.stack([x] * d, axis=0)
    integration = legendre_quadrature(stack_f, integration_bound_a, integration_bound_b)
    assert integration.shape == (d,) + np.broadcast_shapes(integration_bound_a.shape, integration_bound_b.shape)
    assert integration == approx(stack_f(expected_intg(integration_bound_a, integration_bound_b)))



from typing import Literal, Self, overload

import numpy as np
from _typeshed import Incomplete
from numpy.typing import NDArray as NDArray

from relife.data import LifetimeData as LifetimeData
from relife.lifetime_model._plot import PlotECDF as PlotECDF
from relife.lifetime_model._plot import PlotKaplanMeier as PlotKaplanMeier
from relife.lifetime_model._plot import PlotNelsonAalen as PlotNelsonAalen
from relife.lifetime_model._plot import PlotTurnbull as PlotTurnbull

from ._base import NonParametricLifetimeModel as NonParametricLifetimeModel

class ECDF(NonParametricLifetimeModel):
    def __init__(self) -> None: ...
    def fit(
        self,
        time: NDArray[np.float64],
        /,
        event: NDArray[np.float64] | None = None,
        entry: NDArray[np.float64] | None = None,
        departure: NDArray[np.float64] | None = None,
    ) -> Self: ...
    @overload
    def sf(self, se: Literal[False] = False) -> tuple[NDArray[np.float64], NDArray[np.float64]] | None: ...
    @overload
    def sf(
        self, se: Literal[True] = True
    ) -> tuple[NDArray[np.float64], NDArray[np.float64], NDArray[np.float64]] | None: ...
    @overload
    def cdf(self, se: Literal[False] = False) -> tuple[NDArray[np.float64], NDArray[np.float64]] | None: ...
    @overload
    def cdf(
        self, se: Literal[True] = True
    ) -> tuple[NDArray[np.float64], NDArray[np.float64], NDArray[np.float64]] | None: ...
    @property
    def plot(self) -> PlotECDF: ...

class KaplanMeier(NonParametricLifetimeModel):
    def __init__(self) -> None: ...
    def fit(
        self,
        time: NDArray[np.float64],
        event: NDArray[np.float64] | None = None,
        entry: NDArray[np.float64] | None = None,
        departure: NDArray[np.float64] | None = None,
    ) -> Self: ...
    @overload
    def sf(self, se: Literal[False] = False) -> tuple[NDArray[np.float64], NDArray[np.float64]] | None: ...
    @overload
    def sf(
        self, se: Literal[True] = True
    ) -> tuple[NDArray[np.float64], NDArray[np.float64], NDArray[np.float64]] | None: ...
    @property
    def plot(self) -> PlotKaplanMeier: ...

class NelsonAalen(NonParametricLifetimeModel):
    def __init__(self) -> None: ...
    def fit(
        self,
        time: NDArray[np.float64],
        event: NDArray[np.float64] | None = None,
        entry: NDArray[np.float64] | None = None,
        departure: NDArray[np.float64] | None = None,
    ) -> Self: ...
    @overload
    def chf(self, se: Literal[False] = False) -> tuple[NDArray[np.float64], NDArray[np.float64]] | None: ...
    @overload
    def chf(
        self, se: Literal[True] = True
    ) -> tuple[NDArray[np.float64], NDArray[np.float64], NDArray[np.float64]] | None: ...
    @property
    def plot(self) -> PlotNelsonAalen: ...

class Turnbull(NonParametricLifetimeModel):
    tol: Incomplete
    lowmem: Incomplete
    def __init__(self, tol: float | None = 0.0001, lowmem: bool | None = False) -> None: ...
    def fit(
        self,
        time: NDArray[np.float64],
        event: NDArray[np.float64] | None = None,
        entry: NDArray[np.float64] | None = None,
        departure: NDArray[np.float64] | None = None,
        inplace: bool = False,
    ) -> Self: ...
    def sf(self) -> tuple[NDArray[np.float64], NDArray[np.float64]] | None: ...
    @property
    def plot(self) -> PlotTurnbull: ...



import numpy as np
import pytest

from relife.data import NHPPData


@pytest.fixture
def nhpp_data_v0():
    return {
        "ages_at_events": (11, 13, 21, 25, 27),
        "events_assets_ids": ("AB2", "CX13", "AB2", "AB2", "CX13"),
    }


@pytest.fixture
def nhpp_data_v1():
    return {
        "ages_at_events": (11, 13, 21, 25, 27),
        "events_assets_ids": ("AB2", "CX13", "AB2", "AB2", "CX13"),
        "first_ages": (10, 12),
        "last_ages": (35, 60),
        "model_args": (np.array([1.2, 5.5]), np.array([37.2, 22.2])),
        "assets_ids": ("AB2", "CX13"),
    }


@pytest.fixture
def nhpp_data_v2():
    return {
        "ages_at_events": (11, 13, 21, 25, 27),
        "events_assets_ids": ("AB2", "CX13", "AB2", "AB2", "CX13"),
        "model_args": (np.array([1.2, 5.5]), np.array([37.2, 22.2])),
        "assets_ids": ("AB2", "CX13"),
    }


@pytest.fixture
def nhpp_data_v3():
    return {
        "events_assets_ids": ("AB2", "CX13", "AB2", "AB2", "CX13"),
        "assets_ids": ("AB2", "CX13"),
        "first_ages": (10, 12),
        "model_args": (np.array([1.2, 5.5]), np.array([37.2, 22.2])),
        "ages_at_events": (11, 13, 21, 25, 27),
    }


@pytest.fixture
def nhpp_data_v4():
    return {
        "ages_at_events": (11, 13, 21, 25, 27),
        "events_assets_ids": ("AB2", "CX13", "AB2", "AB2", "CX13"),
        "last_ages": (35, 60),
        "model_args": (np.array([1.2, 5.5]), np.array([37.2, 22.2])),
        "assets_ids": ("AB2", "CX13"),
    }


def test_nhhp_data_v0(nhpp_data_v0):
    time, event, entry, model_args = NHPPData(**nhpp_data_v0).to_lifetime_data()
    assert np.all(time == np.array([11.0, 21.0, 25.0, 13.0, 27.0]))
    assert np.all(event == np.array([True, True, True, True, True]))
    assert np.all(entry == np.array([0.0, 11.0, 21.0, 0.0, 13.0]))
    assert model_args == ()


def test_nhhp_data_v1(nhpp_data_v1):
    time, event, entry, model_args = NHPPData(**nhpp_data_v1).to_lifetime_data()
    assert np.all(time == np.array([11.0, 21.0, 25.0, 35.0, 13.0, 27.0, 60.0]))
    assert np.all(event == np.array([True, True, True, False, True, True, False]))
    assert np.all(entry == np.array([10.0, 11.0, 21.0, 25.0, 12.0, 13.0, 27.0]))
    assert np.all(model_args[0] == np.array([1.2, 1.2, 1.2, 1.2, 5.5, 5.5, 5.5]))
    assert np.all(model_args[1] == np.array([37.2, 37.2, 37.2, 37.2, 22.2, 22.2, 22.2]))


def test_nhhp_data_v2(nhpp_data_v2):
    time, event, entry, model_args = NHPPData(**nhpp_data_v2).to_lifetime_data()
    assert np.all(time == np.array([11.0, 21.0, 25.0, 13.0, 27.0]))
    assert np.all(event == np.array([True, True, True, True, True]))
    assert np.all(entry == np.array([0.0, 11.0, 21.0, 0.0, 13.0]))
    assert np.all(model_args[0] == np.array([1.2, 1.2, 1.2, 5.5, 5.5]))
    assert np.all(model_args[1] == np.array([37.2, 37.2, 37.2, 22.2, 22.2]))


def test_nhhp_data_v3(nhpp_data_v3):
    time, event, entry, model_args = NHPPData(**nhpp_data_v3).to_lifetime_data()
    assert np.all(time == np.array([11.0, 21.0, 25.0, 13.0, 27.0]))
    assert np.all(event == np.array([True, True, True, True, True]))
    assert np.all(entry == np.array([10.0, 11.0, 21.0, 12.0, 13.0]))
    assert np.all(model_args[0] == np.array([1.2, 1.2, 1.2, 5.5, 5.5]))
    assert np.all(model_args[1] == np.array([37.2, 37.2, 37.2, 22.2, 22.2]))


def test_nhhp_data_v4(nhpp_data_v4):
    time, event, entry, model_args = NHPPData(**nhpp_data_v4).to_lifetime_data()
    assert np.all(time == np.array([11.0, 21.0, 25.0, 35.0, 13.0, 27.0, 60.0]))
    assert np.all(event == np.array([True, True, True, False, True, True, False]))
    assert np.all(entry == np.array([0.0, 11.0, 21.0, 25.0, 0.0, 13.0, 27.0]))
    assert np.all(model_args[0] == np.array([1.2, 1.2, 1.2, 1.2, 5.5, 5.5, 5.5]))
    assert np.all(model_args[1] == np.array([37.2, 37.2, 37.2, 37.2, 22.2, 22.2, 22.2]))


#
# DATA_V0 = {
#     "assets_ids": np.array(
#         ["S00", "S00", "S01", "S01", "S02", "S11", "S12", "S12", "S13"], dtype=np.str_
#     ),
#     "timeline": np.array(
#         [
#             5.01409347,
#             8.0,
#             4.0485904,
#             8.0,
#             4.0,
#             8.0,
#             5.07210737,
#             8.0,
#             4.0,
#         ],
#         dtype=np.float64,
#     ),
#     "ages": np.array(
#         [5.01409347, 8.0, 4.0485904, 8.0, 4.0, 8.0, 5.07210737, 8.0, 4.0],
#         dtype=np.float64,
#     ),
#     "events_indicators": np.array(
#         [True, False, True, False, False, False, True, False, False], dtype=np.bool_
#     ),
#     "entries": np.array(
#         [2.0, 5.01409347, 0.0, 4.0485904, 0.0, 2.0, 0.0, 5.07210737, 0.0],
#         dtype=np.float64,
#     ),
# }
#
# @pytest.mark.skip(reason="no way of currently testing this")
# def _test_nhpp_to_failure_data(data):
#
#     timeline = data["timeline"]
#     assets_ids = data["assets_ids"]
#     ages = data["ages"]
#     entries = data["entries"]
#     events_indicators = data["events_indicators"]
#
#     sort_ind = np.lexsort((timeline, assets_ids))
#
#     entries = entries[sort_ind]
#     events_indicators = events_indicators[sort_ind]
#     ages = ages[sort_ind]
#     assets_ids = assets_ids[sort_ind]
#
#     first_ages_index = np.roll(assets_ids, 1) != assets_ids
#     last_ages_index = np.roll(first_ages_index, -1)
#
#     immediatly_replaced = np.logical_and(~events_indicators, first_ages_index)
#
#     prefix = np.full_like(assets_ids[immediatly_replaced], "Z", dtype=np.str_)
#     _assets_ids = np.char.add(prefix, assets_ids[immediatly_replaced])
#     first_ages = entries[immediatly_replaced].copy()
#     last_ages = ages[immediatly_replaced].copy()
#
#     events_assets_ids = assets_ids[events_indicators]
#     events_ages = ages[events_indicators]
#     other_assets_ids = np.unique(events_assets_ids)
#     _assets_ids = np.concatenate((_assets_ids, other_assets_ids))
#     first_ages = np.concatenate(
#         (first_ages, entries[first_ages_index & events_indicators])
#     )
#     last_ages = np.concatenate(
#         (last_ages, ages[last_ages_index & ~immediatly_replaced])
#     )
#
#     return events_assets_ids, events_ages, _assets_ids, first_ages, last_ages
#
# @pytest.mark.skip(reason="no way of currently testing this")
# def test_nhpp_to_failure_data():
#
#     events_assets_ids, events_ages, assets_ids, first_ages, last_ages = (
#         _test_nhpp_to_failure_data(DATA_V0)
#     )
#
#     assert np.all(events_assets_ids == np.array(["S00", "S01", "S12"], dtype=np.str_))
#     assert np.all(
#         events_ages == np.array([5.01409347, 4.0485904, 5.07210737], dtype=np.float64)
#     )
#     assert np.all(
#         assets_ids
#         == np.array(["ZS02", "ZS11", "ZS13", "S00", "S01", "S12"], dtype=np.str_)
#     )
#     assert np.all(
#         first_ages == np.array([0.0, 2.0, 0.0, 2.0, 0.0, 0.0], dtype=np.float64)
#     )
#     assert np.all(
#         last_ages == np.array([4.0, 8.0, 4.0, 8.0, 8.0, 8.0], dtype=np.float64)
#     )



import numpy as np
from pytest import approx
from scipy.stats import boxcox, zscore

from relife.lifetime_model import AcceleratedFailureTime, ProportionalHazard, Weibull
from relife.lifetime_model.regression import CovarEffect


def test_covar_effect():
    """
    covar : () or (nb_coef,)
    => g : ()
    => jac_g : (nb_coef,)

    covar : (m, nb_coef)
    => g : (m, 1)
    => jac_g : (nb_coef, m, 1)
    """

    covar_effect = CovarEffect(coefficients=(2.4, 5.5))
    z1 = np.array([1, 2, 3])
    z2 = np.array([0.8, 0.7, 0.5])
    assert covar_effect.g(np.column_stack((z1, z2))) == approx(np.exp(2.4 * z1 + 5.5 * z2).reshape(-1, 1))
    assert covar_effect.jac_g(np.column_stack((z1, z2)))[0] == approx((z1 * np.exp(2.4 * z1 + 5.5 * z2)).reshape(-1, 1))
    assert covar_effect.jac_g(np.column_stack((z1, z2)))[1] == approx((z2 * np.exp(2.4 * z1 + 5.5 * z2)).reshape(-1, 1))

    assert covar_effect.g(np.ones(covar_effect.nb_coef)).shape == ()
    assert covar_effect.g(np.ones((1, covar_effect.nb_coef))).shape == (1, 1)
    assert covar_effect.g(np.ones((10, covar_effect.nb_coef))).shape == (10, 1)

    assert covar_effect.jac_g(np.ones(covar_effect.nb_coef), asarray=True).shape == (covar_effect.nb_coef,)
    assert covar_effect.jac_g(np.ones(covar_effect.nb_coef), asarray=True).shape == (covar_effect.nb_coef,)
    assert covar_effect.jac_g(np.ones((1, covar_effect.nb_coef)), asarray=True).shape == (covar_effect.nb_coef, 1, 1)
    assert covar_effect.jac_g(np.ones((10, covar_effect.nb_coef)), asarray=True).shape == (covar_effect.nb_coef, 10, 1)


def test_rvs(regression, rvs_size, covar, expected_out_shape):
    assert regression.rvs(rvs_size, covar).shape == expected_out_shape(covar=covar, size=rvs_size)
    assert all(
        arr.shape == expected_out_shape(covar=covar, size=rvs_size)
        for arr in regression.rvs(rvs_size, covar, return_event=True)
    )
    assert all(
        arr.shape == expected_out_shape(covar=covar, size=rvs_size)
        for arr in regression.rvs(rvs_size, covar, return_entry=True)
    )
    assert all(
        arr.shape == expected_out_shape(covar=covar, size=rvs_size)
        for arr in regression.rvs(rvs_size, covar, return_event=True, return_entry=True)
    )


def test_sf(regression, time, covar, expected_out_shape):
    assert regression.sf(time, covar).shape == expected_out_shape(time=time, covar=covar)


def test_hf(regression, time, covar, expected_out_shape):
    assert regression.hf(time, covar).shape == expected_out_shape(time=time, covar=covar)


def test_chf(regression, time, covar, expected_out_shape):
    assert regression.chf(time, covar).shape == expected_out_shape(time=time, covar=covar)


def test_cdf(regression, time, covar, expected_out_shape):
    assert regression.cdf(time, covar).shape == expected_out_shape(time=time, covar=covar)


def test_pdf(regression, time, covar, expected_out_shape):
    assert regression.pdf(time, covar).shape == expected_out_shape(time=time, covar=covar)


def test_ppf(regression, probability, covar, expected_out_shape):
    assert regression.ppf(probability, covar).shape == expected_out_shape(time=probability, covar=covar)


def test_ichf(regression, probability, covar, expected_out_shape):
    assert regression.ichf(probability, covar).shape == expected_out_shape(time=probability, covar=covar)


def test_isf(regression, probability, covar, expected_out_shape):
    assert regression.isf(probability, covar).shape == expected_out_shape(time=probability, covar=covar)
    assert regression.isf(np.full(probability.shape, 0.5), covar) == approx(
        np.broadcast_to(regression.median(covar), expected_out_shape(time=probability, covar=covar))
    )


def test_dhf(regression, time, covar, expected_out_shape):
    assert regression.dhf(time, covar).shape == expected_out_shape(time=time, covar=covar)


def test_jac_sf(regression, time, covar, expected_out_shape):
    assert regression.jac_sf(time, covar, asarray=True).shape == (regression.nb_params,) + expected_out_shape(
        time=time, covar=covar
    )


def test_jac_hf(regression, time, covar, expected_out_shape):
    assert regression.jac_hf(time, covar, asarray=True).shape == (regression.nb_params,) + expected_out_shape(
        time=time, covar=covar
    )


def test_jac_chf(regression, time, covar, expected_out_shape):
    assert regression.jac_chf(time, covar, asarray=True).shape == (regression.nb_params,) + expected_out_shape(
        time=time, covar=covar
    )


def test_jac_cdf(regression, time, covar, expected_out_shape):
    assert regression.jac_cdf(time, covar, asarray=True).shape == (regression.nb_params,) + expected_out_shape(
        time=time, covar=covar
    )


def test_jac_pdf(regression, time, covar, expected_out_shape):
    assert regression.jac_pdf(time, covar, asarray=True).shape == (regression.nb_params,) + expected_out_shape(
        time=time, covar=covar
    )


def test_ls_integrate(regression, integration_bound_a, integration_bound_b, covar, expected_out_shape):
    # integral_a^b dF(x)
    integration = regression.ls_integrate(np.ones_like, integration_bound_a, integration_bound_b, covar, deg=100)
    assert integration.shape == expected_out_shape(a=integration_bound_a, b=integration_bound_b, covar=covar)
    assert integration == approx(
        regression.cdf(integration_bound_b, covar) - regression.cdf(integration_bound_a, covar)
    )
    # integral_0^inf x*dF(x)
    integration = regression.ls_integrate(
        lambda x: x, np.zeros_like(integration_bound_a), np.full_like(integration_bound_b, np.inf), covar, deg=100
    )
    assert integration == approx(
        np.broadcast_to(
            regression.mean(covar),
            expected_out_shape(a=integration_bound_a, b=integration_bound_b, covar=covar),
        ),
        rel=1e-3,
    )


def test_aft_pph_weibull_eq(insulator_string_data):
    covar = zscore(
        np.column_stack(
            (
                boxcox(insulator_string_data["pHCl"])[0],
                boxcox(insulator_string_data["pH2SO4"])[0],
                boxcox(insulator_string_data["HNO3"])[0],
            )
        )
    )
    weibull_aft = AcceleratedFailureTime(Weibull()).fit(
        insulator_string_data["time"],
        covar,
        event=insulator_string_data["event"],
        entry=insulator_string_data["entry"],
    )
    weibull_pph = ProportionalHazard(Weibull()).fit(
        insulator_string_data["time"],
        covar,
        event=insulator_string_data["event"],
        entry=insulator_string_data["entry"],
    )

    assert weibull_pph.baseline.params == approx(weibull_aft.baseline.params, rel=1e-3)
    assert weibull_pph.covar_effect.params == approx(
        -weibull_aft.baseline.shape * weibull_aft.covar_effect.params,
        rel=1e-3,
    )



from __future__ import annotations

from typing import Generic, Optional, TypedDict, TypeVar

import numpy as np
from numpy.typing import NDArray

from relife.economic import ExponentialDiscounting, Reward, cost
from relife.lifetime_model import (
    FrozenAgeReplacementModel,
    FrozenLeftTruncatedModel,
)
from relife.lifetime_model.distribution import LifetimeDistribution
from relife.lifetime_model.regression import FrozenLifetimeRegression
from relife.stochastic_process import RenewalRewardProcess
from relife.stochastic_process._sample.data import RenewalRewardProcessSample

M = TypeVar("M", LifetimeDistribution, FrozenLifetimeRegression, FrozenAgeReplacementModel, FrozenLeftTruncatedModel)
R = TypeVar("R", bound=Reward)


class LifetimeFitArg(TypedDict):
    time: NDArray[np.float64]
    event: NDArray[np.bool_]
    entry: NDArray[np.float64]
    args: tuple[NDArray[np.float64], ...]


class BaseOneCycleAgeReplacementPolicy(Generic[M, R]):

    def __init__(
        self,
        lifetime_model: M,
        reward: R,
        discounting_rate: float = 0.0,
        period_before_discounting: float = 1.0,
    ) -> None:
        self.cost = cost
        self.reward = reward
        self.discounting = ExponentialDiscounting(discounting_rate)
        if period_before_discounting <= 0:
            raise ValueError("The period_before_discounting must be greater than 0")
        self.period_before_discounting = period_before_discounting
        self.lifetime_model = lifetime_model
        self.count_data = None

    @property
    def discounting_rate(self):
        return self.discounting.rate

    def _make_timeline(self, tf: float, nb_steps: int) -> NDArray[np.float64]:
        # tile is necessary to ensure broadcasting of the operations
        timeline = np.linspace(0, tf, nb_steps, dtype=np.float64)  # (nb_steps,)
        args_nb_assets = getattr(self.lifetime_model, "nb_assets", 1)  # default 1 for LifetimeDistribution case
        if args_nb_assets > 1:
            timeline = np.tile(timeline, (args_nb_assets, 1))
        elif self.reward.ndim == 2:  # elif because we consider that if m > 1 in frozen_model, in reward it is 1 or m
            timeline = np.tile(timeline, (self.reward.size, 1))
        return timeline  # (nb_steps,) or (m, nb_steps)w

    def expected_total_cost(self, tf: float, nb_steps: int) -> tuple[NDArray[np.float64], NDArray[np.float64]]:
        timeline = self._make_timeline(tf, nb_steps)
        etc = self.lifetime_model.ls_integrate(
            lambda x: self.reward.conditional_expectation(x) * self.discounting.factor(x),
            np.zeros_like(timeline),
            timeline,
            deg=15,
        )  # (nb_steps,) or (m, nb_steps)
        if timeline.ndim == 2:
            return timeline[0, :], etc  # (nb_steps,) and (m, nb_steps)
        return timeline, etc  # (nb_steps,) and (nb_steps,)

    def asymptotic_expected_total_cost(self) -> NDArray[np.float64]:
        # reward partial expectation
        return np.squeeze(
            self.lifetime_model.ls_integrate(
                lambda x: self.reward.conditional_expectation(x) * self.discounting.factor(x), 0.0, np.inf, deg=15
            )
        )  # () or (m,)

    def _expected_equivalent_annual_cost(
        self, timeline: NDArray[np.float64]
    ) -> tuple[NDArray[np.float64], NDArray[np.float64]]:

        # timeline : (nb_steps,) or (m, nb_steps)
        def f(x: float | NDArray[np.float64]) -> tuple[NDArray[np.float64], NDArray[np.float64]]:
            # avoid zero division + 1e-6
            return (
                self.reward.conditional_expectation(x)
                * self.discounting.factor(x)
                / (self.discounting.annuity_factor(x) + 1e-6)
            )

        q0 = self.lifetime_model.cdf(self.period_before_discounting) * f(self.period_before_discounting)  # () or (m, 1)
        a = np.full_like(timeline, self.period_before_discounting)  # (nb_steps,) or (m, nb_steps)

        # change first value of lower bound to compute the integral
        a[timeline < self.period_before_discounting] = 0.0  # (nb_steps,)
        # a = np.where(timeline < self.period_before_discounting, 0., a)  # (nb_steps,)
        integral = self.lifetime_model.ls_integrate(
            f, a, timeline, deg=15
        )  # (nb_steps,) or (m, nb_steps) if q0: (), or (m, nb_steps) if q0 : (m, 1)
        mask = np.broadcast_to(
            timeline < self.period_before_discounting, integral.shape
        )  # (), (nb_steps,) or (m, nb_steps)
        q0 = np.broadcast_to(q0, integral.shape)  # (nb_steps,) or (m, nb_steps)
        integral = np.where(mask, q0, q0 + integral)
        if timeline.ndim == 2:
            return timeline[0, :], integral  # (nb_steps,) and (m, nb_steps)
        return timeline, integral  # (nb_steps,) and (nb_steps,)

    def expected_equivalent_annual_cost(
        self, tf: float, nb_steps: int
    ) -> tuple[NDArray[np.float64], NDArray[np.float64]]:
        timeline = self._make_timeline(tf, nb_steps)  # (nb_steps,) or (m, nb_steps)
        return self._expected_equivalent_annual_cost(timeline)  # (nb_steps,) or (m, nb_steps)

    def asymptotic_expected_equivalent_annual_cost(
        self,
    ) -> NDArray[np.float64]:
        timeline = np.array(np.inf)
        args_nb_assets = getattr(self.lifetime_model, "args_nb_assets", 1)  #  default 1 for LifetimeDistribution case
        if args_nb_assets > 1:
            timeline = np.tile(timeline, (args_nb_assets, 1))
        elif self.reward.ndim == 2:  # elif because we consider that if m > 1 in frozen_model, in reward it is 1 or m
            timeline = np.tile(timeline, (self.reward.size, 1))
        # timeline : () or (m, 1)
        return np.squeeze(self._expected_equivalent_annual_cost(timeline)[-1])  # () or (m,)


# TODO : generic of generic
class BaseAgeReplacementPolicy(Generic[M, R]):

    def __init__(self, stochastic_process: RenewalRewardProcess[M, R]):
        self.stochastic_process = stochastic_process

    @property
    def discounting_rate(self):
        return self.stochastic_process.discounting_rate

    @discounting_rate.setter
    def discounting_rate(self, value: float) -> None:
        self.stochastic_process.discounting_rate = value

    def expected_nb_replacements(self, tf: float, nb_steps: int) -> tuple[NDArray[np.float64], NDArray[np.float64]]:
        r"""
        The expected number of replacements.

        It is computed  by solving the renewal equation:

        .. math::

            m(t) = F_1(t) + \int_0^t m(t-x) \mathrm{d}F(x)

        where:

        - :math:`m` is the renewal function,
        - :math:`F` is the cumulative distribution function of the underlying
          lifetime model,
        - :math:`F_1` is the cumulative distribution function of the underlying
          lifetime model for the fist renewal in the case of a delayed renewal
          process.


        Parameters
        ----------
        tf : float
            Time horizon. The expected number of replacements will be computed up until this calendar time.
        nb_steps : int
            The number of steps used to compute the expected number of replacements

        Returns
        -------
        tuple of two ndarrays
            A tuple containing the timeline used to compute the expected number of replacements and its corresponding values at each
            step of the timeline.
        """
        return self.stochastic_process.renewal_function(tf, nb_steps)  # (nb_steps,) or (m, nb_steps)

    def expected_total_cost(self, tf: float, nb_steps: int) -> tuple[NDArray[np.float64], NDArray[np.float64]]:
        r"""
        The expected total cost.

        It is computed by solving the renewal equation and is given by:

        .. math::

            z(t) = \mathbb{E}(Z_t) = \int_{0}^{\infty}\mathbb{E}(Z_t~|~X_1 = x)dF(x)

        where :

        - :math:`t` is the time
        - :math:`X_i \sim F` are :math:`n` random variable lifetimes, *i.i.d.*, of cumulative distribution :math:`F`.
        - :math:`Z_t` is the random variable reward at each time :math:`t`.
        - :math:`\delta` is the discounting rate.

        Parameters
        ----------
        tf : float
            Time horizon. The expected total cost will be computed up until this calendar time.
        nb_steps : int
            The number of steps used to compute the expected total cost

        Returns
        -------
        tuple of two ndarrays
            A tuple containing the timeline used to compute the expected total cost and its corresponding values at each
            step of the timeline.
        """
        return self.stochastic_process.expected_total_reward(tf, nb_steps)  # (nb_steps,) or (m, nb_steps)

    def asymptotic_expected_total_cost(self) -> NDArray[np.float64]:
        r"""
        The asymtotic expected total cost

        .. math::

            \lim_{t\to\infty} z(t)

        where :math:`z(t)` is the expected total cost at :math:`t`. See :py:meth:`~AgeReplacementPolicy.expected_total_cost` for more details.

        Returns
        -------
        ndarray
            The asymptotic expected total cost values
        """
        return self.stochastic_process.asymptotic_expected_total_reward()  # () or (m, 1)

    def expected_equivalent_annual_cost(
        self, tf: float, nb_steps: int
    ) -> tuple[NDArray[np.float64], NDArray[np.float64]]:
        r"""
        The expected equivalent annual cost.

        .. math::

            \text{EEAC}(t) = \dfrac{\delta z(t)}{1 - e^{-\delta t}}

        where :

        - :math:`t` is the time
        - :math:`z(t)` is the expected_total_cost at :math:`t`. See :py:meth:`~AgeReplacementPolicy.expected_total_cost` for more details.`.
        - :math:`\delta` is the discounting rate.

        Parameters
        ----------
        tf : float
            Time horizon. The expected equivalent annual cost will be computed up until this calendar time.
        nb_steps : int
            The number of steps used to compute the expected equivalent annual cost

        Returns
        -------
        tuple of two ndarrays
            A tuple containing the timeline used to compute the expected annual cost and its corresponding values at each
            step of the timeline.
        """
        return self.stochastic_process.expected_equivalent_annual_worth(tf, nb_steps)  # (nb_steps,) or (m, nb_steps)

    def asymptotic_expected_equivalent_annual_cost(self) -> NDArray[np.float64]:
        r"""
        The asymtotic expected equivalent annual cost

        .. math::

            \lim_{t\to\infty} \text{EEAC}(t)

        where :math:`\text{EEAC}(t)` is the expected equivalent annual cost at :math:`t`. See :py:meth:`~AgeReplacementPolicy.expected_equivalent_annual_cost` for more details.

        Returns
        -------
        ndarray
            The asymptotic expected equivalent annual cost
        """
        return self.stochastic_process.asymptotic_expected_equivalent_annual_worth()  # () or (m, 1)

    def sample(
        self,
        size: int,
        tf: float,
        t0: float = 0.0,
        seed: Optional[int] = None,
    ) -> RenewalRewardProcessSample:
        """Renewal data sampling.

        This function will sample data and encapsulate them in an object.

        Parameters
        ----------
        size : int
            The size of the desired sample.
        tf : float
            Time at the end of the observation.
        t0 : float, default 0
            Time at the beginning of the observation.
        size : int or tuple of 2 int
            Size of the sample
        seed : int, optional
            Random seed, by default None.

        """

        return self.stochastic_process.sample(tf, t0, size, seed)

    def generate_lifetime_data(
        self,
        size: int,
        tf: float,
        t0: float = 0.0,
        seed: Optional[int] = None,
    ) -> LifetimeFitArg:
        """Generate lifetime data

        This function will generate lifetime data that can be used to fit a lifetime model.

        Parameters
        ----------
        size : int
            The size of the desired sample.
        tf : float
            Time at the end of the observation.
        t0 : float, default 0
            Time at the beginning of the observation.
        seed : int, optional
            Random seed, by default None.

        Returns
        -------
        A dict of time, event, entry and args (covariates)

        """
        return self.stochastic_process.generate_lifetime_data(tf, t0, size, seed)



from pathlib import Path

import numpy as np
from numpy.typing import NDArray


def load_power_transformer() -> NDArray[np.void]:
    r"""
    Load power transformer dataset containing observed lifetimes and left truncations values.

    Examples
    --------
    >>> from relife.data import load_power_transformer
    >>> data = load_power_transformer()
    >>> print(data["time"])
    [34.3 45.1 53.2 ... 30.  30.  30. ]

    Returns
    -------
    structured array
        A numpy structured array of 3 fields :

        - time (``np.float64``) : observed lifetime values
        - event (``np.bool_``) : boolean flag indicated if the event has been observed or not (if False, the observed lifetimes are right censored)
        - entry (``np.float64``) : left truncation values

    """

    data = np.loadtxt(
        Path(Path(__file__).parents[0], "csv/power_transformer.csv"),
        delimiter=",",
        skiprows=1,
        dtype=np.dtype([("time", np.float64), ("event", np.float64), ("entry", np.float64)]),
    )
    # for some reason, numpy can't cast 1.0/0.0 to np.bool_
    new_dtype = np.dtype([("time", np.float64), ("event", np.bool_), ("entry", np.float64)])
    return data.astype(new_dtype)


def load_insulator_string() -> NDArray[np.void]:
    r"""
    Load insulator string dataset containing observed lifetimes, left truncations values and covariates

    Examples
    --------
    >>> from relife.data import load_insulator_string
    >>> data = load_insulator_string()
    >>> print(data["time"])
    [70.  30.  45.  ...  8.8  7.6 53. ]
    >>> print(data["pHCl"])
    [0.49 0.76 0.43 ... 1.12 1.19 0.35]

    Returns
    -------
    structured array
        A numpy structured array of 3 fields :

        - time (``np.float64``) : observed lifetime values
        - event (``np.bool_``) : boolean flag indicated if the event has been observed or not (if False, the observed lifetimes are right censored)
        - entry (``np.float64``) : left truncation values
        - pHCl (``np.float64``) : quantitative covariate values (concentration of pHCl)
        - pH2SO4 (``np.float64``) : quantitative covariate values (concentration of pH2SO4)
        - HNO3 (``np.float64``) : quantitative covariate values (concentration of HNO3)

    """

    data = np.loadtxt(
        Path(Path(__file__).parents[0], "csv/insulator_string.csv"),
        delimiter=",",
        skiprows=1,
        dtype=np.dtype(
            [
                ("time", np.float64),
                ("event", np.float64),
                ("entry", np.float64),
                ("pHCl", np.float64),
                ("pH2SO4", np.float64),
                ("HNO3", np.float64),
            ]
        ),
    )
    # for some reason, numpy can't cast 1.0/0.0 to np.bool_
    new_dtype = np.dtype(
        [
            ("time", np.float64),
            ("event", np.bool_),
            ("entry", np.float64),
            ("pHCl", np.float64),
            ("pH2SO4", np.float64),
            ("HNO3", np.float64),
        ]
    )
    return data.astype(new_dtype)


def load_circuit_breaker() -> NDArray[np.void]:
    r"""
    Load circuit breaker dataset containing observed lifetimes and left truncations values

    Examples
    --------
    >>> from relife.data import load_circuit_breaker
    >>> data = load_circuit_breaker()
    >>> print(data["time"])
    [34. 28. 12. ... 42. 42. 37.]

    Returns
    -------
    structured array
        A numpy structured array of 3 fields :

        - time (``np.float64``) : observed lifetime values
        - event (``np.bool_``) : boolean flag indicated if the event has been observed or not (if False, the observed lifetimes are right censored)
        - entry (``np.float64``) : left truncation values
    """

    data = np.loadtxt(
        Path(Path(__file__).parents[0], "csv/circuit_breaker.csv"),
        delimiter=",",
        skiprows=1,
        dtype=np.dtype([("time", np.float64), ("event", np.float64), ("entry", np.float64)]),
    )
    # for some reason, numpy can't cast 1.0/0.0 to np.bool_
    new_dtype = np.dtype([("time", np.float64), ("event", np.bool_), ("entry", np.float64)])
    return data.astype(new_dtype)


# def load_input_turnbull() -> NDArray[np.void]:
#     """_summary_
#
#     Returns:
#         np.ndarray: _description_
#     """
#     data = np.loadtxt(
#         Path(Path(__file__).parents[0], "csv/input_turnbull.csv"),
#         delimiter=",",
#         skiprows=1,
#     )
#     data["event"] = data["event"].astype(np.bool_)
#     return data



from .base import StochasticProcess as StochasticProcess
from .non_homogeneous_poisson_process import (
    FrozenNonHomogeneousPoissonProcess as FrozenNonHomogeneousPoissonProcess,
)
from .non_homogeneous_poisson_process import (
    NonHomogeneousPoissonProcess as NonHomogeneousPoissonProcess,
)
from .renewal_process import RenewalProcess as RenewalProcess
from .renewal_process import RenewalRewardProcess as RenewalRewardProcess
